;;;; The XML-in-LAML common parts, which are shared between all XML-in-LAML languages.
;;;; This library is designed to co-exists with non-validating HTML mirrors (for the sake of LENO, for instance).
;;;; It cannot be used together with validating ast-based HTML mirrors, such as html4.01-transitional-validating.
;;;; If you need to, you should use an all XML-IN-LAML solution. By that we mean a solution, where also the
;;;; HTML stuff is based on XML-in-LAML. In practical terms, it means that you should use one of the XHTML mirrors in
;;;; lib/xml-in-laml/mirrors/.<p>
;;;;
;;;; This library loads <a href="../../man/final-state-automaton.html">the the LAML final state automation library</a>
;;;; for the sake of full document validation at document generation time.

(lib-load "final-state-automaton.scm")

;;; XML front matter stuff.
;;; This section contains the XML declaration and other XML front matter stuff

;; The standard-prolog function as redefined for XML. It returns the xml-declaration,
;; the document type declaration, and a copyright comment.
;; The standard prolog is the document part before the document root element.
;; A default version of the standard-prolog is found in laml.scm. 
;; The present function relies on another function, xml-document-type-declaration-in, which returns
;; an appropriate document type declaration
;; .form (standard-prolog [language])
;; .internal-references "applied function" "xml-declaration"
;; .internal-references "applied function" "xml-document-type-declaration-in"
;; .reference "applied function" "copyright-clause" "../../../man/laml.html#copyright-clause"
(define (standard-prolog . optional-parameter-list)
 (let* ((language (optional-parameter 1 optional-parameter-list #f))
        (doc-type-decl (xml-document-type-declaration-in language))
       )
  (string-append
   (xml-declaration) (as-string #\newline)
   doc-type-decl
   (if (not (empty-string? doc-type-decl)) (as-string #\newline) "")
   (copyright-clause)
   (if (not (empty-string? (copyright-clause))) (as-string #\newline) ""))))


;; Return the xml declaration
(define (xml-declaration)
  "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>")

;;; XML-in-LAML processing parameters.
;;; The variables and functions in this part controls a number of general properties of XML-in-LAML processing. Some of these
;;; are generic for all XML-in-LAML languages; Others are specific to a single language.
;;; A number of the properties are related to the checking, validation, and error reporting of the XML mirror functions.

;; A boolean variable which controls the checking of element attributes.
;; If true, check the element attributes against the attributes as defined in the DTD file.
;; This variable is generic and common for all XML-in-LAML languages.
;; The default value is true.
(define xml-check-attributes? #t)

;; A boolean variable that controls the XML validation.
;; If true, do validate the generated XML document against the DTD.
;; This variable is generic and common for all XML-in-LAML languages.
;; The default value is true.
(define xml-validate-contents? #t)

;; A boolean variable that controls the check of XML language overlap.
;; If true, check that that no elements are used via ambiguous simple names.
;; Quite naturally, this variable is generic and common for XML-in-LAML loaded at the same time. 
(define xml-check-language-overlap? #t)


;; Return if CDATA  (Character data) is transformed through an HTML/XML character transformation table. 
;; The value is boolean, and normally true.
;; .internal-references "mutator" "set-xml-transliterate-character-data-in"
;; .reference "transformation table" "laml.scm" "../../../man/laml.html#html-char-transformation-table"
(define (xml-transliterate-character-data-in? language)
  (assert-known-xml-language language "xml-transliterate-character-data-in?")
  (eval-cur-env (aggregated-variable (as-string language) "xml-transliterate-character-data?")))

;; Ask for transliteration of all CDATA characters via an HTML/XML character transformation table. 
;; new-value must be boolean.
;; .internal-references "selector" "xml-transliterate-character-data-in?"
;; .internal-references "table getter" "xml-char-transformation-table-in"
;; .internal-references "table setter" "set-xml-char-transformation-table-in"
;; .internal-references "exceptions" "xml-non-transliteration-elements-in"
;; .reference "transformation table" "laml.scm" "../../../man/laml.html#html-char-transformation-table"
;; .parameter language The name of the XML-in-LAML language - see the first few linew of relevant mirror library manuals - a symbol.
;; .parameter new-value A boolean value.
;; .misc Consider macro implementation.
(define (set-xml-transliterate-character-data-in language new-value)
  (assert-known-xml-language language "set-xml-transliterate-character-data-in")
  (eval-cur-env (list 'set! (aggregated-variable (as-string language) "xml-transliterate-character-data?") new-value)))

;; Return the HTML/XML character transformation table used for language.
;; .returns A character transformation table.
;; .reference "transformation table" "laml.scm" "../../../man/laml.html#html-char-transformation-table"
;; .internal-references "table setter" "set-xml-char-transformation-table-in"
;; .internal-references "exceptions" "xml-non-transliteration-elements-in"
;; .reference "transformation table" "laml.scm" "../../../man/laml.html#html-char-transformation-table"
(define (xml-char-transformation-table-in language)
  (assert-known-xml-language language "xml-char-transformation-table-in")
  (eval-cur-env (aggregated-variable (as-string language) "xml-char-transformation-table")))

;; Set the HTML/XML character transformation table used for language.
;; The value must be a character transformation table. 
;; .internal-references "table getter" "xml-char-transformation-table-in"
;; .internal-references "exceptions" "xml-non-transliteration-elements-in"
;; .reference "transformation table" "laml.scm" "../../../man/laml.html#html-char-transformation-table"
;; .misc  Consider macro implementation.
;; .parameter language The name of the XML-in-LAML language - see the first few linew of relevant mirror library manuals - a symbol.
;; .parameter new-value A character transformation table.
(define (set-xml-char-transformation-table-in language new-value)
  (assert-known-xml-language language "set-xml-char-transformation-table-in")
  (eval-cur-env (list 'set! (aggregated-variable (as-string language) "xml-char-transformation-table") new-value)))



;; Return the list of element names, for which we do not carry out the character transliteration in language.
;; In XHTML, the list typically include the elements style and script.
;; The value is a list of strings.
;; The value of this function is locked and bound at mirror generation time.
;; .internal-references "transliteration" "xml-transliterate-character-data-in?"
(define (xml-non-transliteration-elements-in language)
  (assert-known-xml-language language "xml-non-transliteration-elements-in")
  (eval-cur-env (aggregated-variable (as-string language) "xml-non-transliteration-elements")))

;; Return the list of element names, for which we consistenly render all white space content characters as given in the input.
;; In XHTML, the list typically include the pre element.
;; The value is a list of strings.
;; The value of this function is locked and bound at mirror generation time.
(define (xml-preformatted-text-elements-in language)
  (assert-known-xml-language language "xml-preformatted-text-elements-in")
  (eval-cur-env (aggregated-variable (as-string language) "xml-preformatted-text-elements")))



;; How many characters of the validation error messages to be presented.
;; If you want longer error messages, just increase the value of this variable.
;; The default value is currently 130.
(define xml-error-truncation-length 130)


;; Controls whether the default attribute values, as specified in the DTD, are passed explicitly as attribute values
;; in every instance of the element. 
;; The value is boolean, and normally false. 
;; A true value in most cases give problems relative to the XML attribute validation.
;; The value of this function is locked and bound at mirror generation time.
;; .internal-references "validation control" "xml-check-attributes?"
(define (xml-pass-default-dtd-attributes-in? language)
  (assert-known-xml-language language "xml-pass-default-dtd-attributes-in?")
  (eval-cur-env (aggregated-variable (as-string language) "xml-pass-default-dtd-attributes?")))


;; A boolean variable that controls how rigid LAML handles attribute values in language (first parameter).
;; As the name indicates, the value controls if only string valued attributes are allowed in the language.
;; If no language information is present (i.e., if the language parameter is #f) always return #t. 
;; .internal-references "setter" "set-xml-accept-only-string-valued-attributes-in"
(define (xml-accept-only-string-valued-attributes-in? language)
  (if language
      (begin
        (assert-known-xml-language language "xml-accept-only-string-valued-attributes-in?")
        (eval-cur-env (aggregated-variable (as-string language) "xml-accept-only-string-valued-attributes?"))
      )
      #t))

;; Controls the handling of LAML attribute values in language.
;; If new-value is true, LAML only accept string valued XML attribute.
;; If new-value is false, LAML string converts whatever follows an attribute symbol.
;; The value must be a boolean value. True is recommended.
;; .internal-references "getter" "xml-accept-only-string-valued-attributes-in?"
;; .misc Consider macro implementation.
;; .parameter language The name of the XML-in-LAML language - see the first few linew of relevant mirror library manuals - a symbol.
;; .parameter new-value A boolean value.
(define (set-xml-accept-only-string-valued-attributes-in language new-value)
  (assert-known-xml-language language "set-xml-accept-only-string-valued-attributes-in")
  (eval-cur-env (list 'set! (aggregated-variable (as-string language) "xml-accept-only-string-valued-attributes?") new-value)))



;; Return the document type declaration of language.
;; The empty string signals that no document type declaration is available.
;; The value of this function is locked and bound at mirror generation time.
(define (xml-document-type-declaration-in language)
  (if language
      (begin
        (assert-known-xml-language language "xml-document-type-declaration-in")
        (eval-cur-env (aggregated-variable (as-string language) "xml-document-type-declaration")))
      ""))


;; This value controls the representation of white space in the internal AST representation of a document.
;; If true, white space markers are inserted. If false, a clean AST whithou any white space marking is produced.
;; In some languages, white space handling is essential. This is the case in XHTML.
;; In non-presentational XML language, white space handling is not relevant.
;; The value of this function is locked and bound at mirror generation time.
(define (xml-represent-white-space-in? language)
  (if language
      (begin
        (assert-known-xml-language language "xml-represent-white-space-in?")
        (eval-cur-env (aggregated-variable (as-string language) "xml-represent-white-space?"))
      )
      #t))


; Form a variable of language-string variable-string and return the symbol.
(define (aggregated-variable language-string variable-string)
  (as-symbol (string-append language-string "-" variable-string)))

; Assert that language is legal relative to the XML-in-LAML languages in use.
; The optional second parameter is intended to hold the operation causing the problems.
; If not, stop the program (fatal error)
(define (assert-known-xml-language language . optional-parameter-list)
 (let ((context (optional-parameter 1 optional-parameter-list #f)))
  (if (not (symbol? language))
      (laml-error (if context (string-append (as-string context) ":") "") "The XML-in-LAML language must be given as a symbol:" language))
  (if (not (memq language (languages-in-use)))
      (laml-error (if context (string-append (as-string context) ":" (as-string #\newline)) "")
                  "The language" language 
                  "is not among the currently loaded XML-in-LAML languages:"
                  (list-to-string (map as-string (languages-in-use)) ", ")))))

;;; Other Constants.

;; The explicit white space value, as used internally in ast. The default value is boolean #t.
(define explicit-space #t)

;; The explicit white space suppress value. The default value is boolean #f.
(define explicit-space-suppress #f)

;; The symbolic name of explicit white space suppress.
(define _ explicit-space-suppress)

;; An integer that expresses the prefered maximum column width for pretty printed ouput.
(define preferred-maximum-width 90)

;; An integer which gives the level of indentation for pretty printed output.
(define indentation-delta 3)




; ---------------------------------------------------------------------------------------------------
;;; AST functions.
;;; In this section we document the functions that work on abstract syntax trees (ASTs).
;;; Abstract syntax trees make up the internal representation of XML documents in LAML.
;;; All the XML mirror functions return abstract syntax trees.

; A boolean variable that controls whether or not we check the internal soundness of ASTs.
; If the basic mirror is used directly, it makes good sense to set this variable to #t.
; Else we recommend a #f value.
(define check-ast-constituents? #f)

;; Make an AST given element-name, contents, a property list of attributes, a kind (single/double)
;; an an XML language name. 
;; The parameter contents may be a list of subtrees. It may also be an element which satisfies
;; either ast?, cdata? or forced-white-space?. In these cases, we embed the single element into a list.
;; attributes is the list of attributes on property list format (such as '(a1 "v1" a2 "v2")). 
;; attribute names are always symbols, and attribute values are normally strings.
;; .parameter element-name A symbol or a string.
;; .parameter contents A list of contents in term of CDATA (strings), white space markers, or other ASTs.
;; .parameter attributes The attributes represented as a property list
;; .parameter kind The symbol double or single
;; .parameter language A symbol or string that represent the XML-in-LAL language name. 
(define (make-ast element-name contents attributes kind language)
 (let ((subtrees (cond ((ast? contents) (list contents))
                       ((cdata? contents) (list contents))
                       ((forced-white-space? contents) (list contents))
                       ((list? contents) contents)
                       (else (laml-error "make-ast: Contents must be ast, cdata, forced white space, or a list of these: " 
                                         (as-string contents))))))
  (list 'ast (as-string element-name) subtrees attributes (as-symbol kind) (as-symbol language))))

; AST selectors.

;; Return the root element name of ast. 
;; The type of the returned value is a string.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
;; .form (ast-element-name ast)
(define ast-element-name (make-selector-function 2 "ast-element-name"))

;; Return the list of substrees of ast.  This is also known as the contents of the ast. It is a list of CDATA (strings), ASTs, and white space markers.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
;; .form (ast-subtrees ast)
(define ast-subtrees (make-selector-function 3 "ast-subtrees"))

;; Return a specific subtree of ast, namely the n'th subtree with element-name el-name.
;; If no such subtree exists, return #f.
;; .pre-condition n > 0
;; .returns An AST (if located) or #f
;; .parameter ast The ast in which to locate a subtree
;; .parameter el-name The ast element name of a subtree of ast (string or symbol)
;; .parameter n The constituent number regarding el-name subtrees (an integer).
;; .form (sub-ast ast el-name [n])
(define (sub-ast ast el-name . optional-parameter-list)
 (let ((n (optional-parameter 1 optional-parameter-list 1))
       (subtrees (ast-subtrees ast)))
  (sub-ast-1 subtrees (as-string el-name) n)))

(define (sub-ast-1 subtree-list el-name n)
 (cond ((null? subtree-list) #f)
       ((and (equal? el-name (ast-element-name (car subtree-list))) (= n 1))
         (car subtree-list))
       ((and (equal? el-name (ast-element-name (car subtree-list))) (> n 1))
         (sub-ast-1 (cdr subtree-list) el-name (- n 1)))
       (else (sub-ast-1 (cdr subtree-list) el-name n))))
   



;; Return the list of attributes of ast, on property list format.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
;; .form (ast-attributes ast)
;; .misc The functions get-prop and defaulted-get-prop are useful for accessing the individual attribute values in the returned property list.
;; .reference "Relevant function" "get-prop" "../../man/general.html#get-prop"
;; .returns Return value description

(define ast-attributes (make-selector-function 4 "ast-attributes"))

;; Return the kind of the ast. 
;; The type of the returned value is a symbol, either single or double.
;; ASTs of type single are also known as empty elements.
;; Tells whether the ast is to be rendered as a single or double tag.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
;; .form (ast-kind ast)
(define ast-kind (make-selector-function 5 "ast-kind"))

;; Return the language of the ast.
;; The type of the returned value is a symbol.
;; .pre-condition ast is an abstract syntax tree that satisfies the ast? predicate
;; .form (ast-language ast)
(define ast-language (make-selector-function 6 "ast-language"))



; AST predicates

;; Is the parameter x an AST
(define (ast? x)
  (and (pair? x) (eq? (car x) 'ast)
       (list? x) (= 6 (length x))))

;; Is the parameter x CDATA
;; .form (cdata? x)
(define cdata? string?)

;; Is the parameter x either an ast or a contents string.
(define (contents-data? x)
 (or (cdata? x) (ast? x) (char-ref? x)))

;; Is the parameter x a white space mark.
(define (forced-white-space? x)
 (eq? x explicit-space))

;; Is the parameter x a white suppress space mark.
(define (white-space-suppress? x)
 (eq? x explicit-space-suppress))

;; Is the parameter x white space related?
(define (white-space-related? x)
 (or (eq? x explicit-space) (eq? x explicit-space-suppress)))

;; Is x an ast nodes without subtrees?
(define (terminal-ast-node? x)
 (and (ast? x)
      (null? (ast-subtrees x))))

;; A higher-order function that returns an AST predicate, which checks if the type of the AST is name.
;; The type is either element-name, kind, or language (a symbol).
;; If type is the symbol element-name, the generated predicate checks if a given AST has that element name at top level.
;; If type is the symbol kind, the generated predicate checks if a given AST is either double or single.
;; If type is the symbol language, the generated predicate checks if a given AST belongs to the language.
(define (ast-of-type? type name)
  (cond ((eq? type 'element-name)
           (lambda (ast) 
	     (equal? (as-string (ast-element-name ast)) (as-string name))))
        ((eq? type 'kind)
           (lambda (ast) 
             (equal? (as-string (ast-kind ast)) (as-string name))))
        ((eq? type 'language)
           (lambda (ast) 
             (equal? (as-string (ast-language ast)) (as-string name))))
        (else 
            (laml-error "ast-of-type?: Unknown first parameter (type): " type))))


;; Return a string of the aggregated CDATA contents of ast.
;; White space markers are taken into account.  AST subtrees are of ast are just ignored.
(define (aggregated-ast-cdata-contents ast)
  (aggregated-ast-cdata-contents-1 (ast-subtrees ast) ""))

(define (aggregated-ast-cdata-contents-1 contents-list res)
  (cond ((null? contents-list) res)
        ((cdata? (car contents-list)) 
           (aggregated-ast-cdata-contents-1 (cdr contents-list) (string-append res (car contents-list))))
        ((forced-white-space? (car contents-list)) (aggregated-ast-cdata-contents-1 (cdr contents-list) (string-append res " ")))
        ((white-space-suppress? (car contents-list)) (aggregated-ast-cdata-contents-1 (cdr contents-list) res))
        ((ast? (car contents-list)) ; drop
           (aggregated-ast-cdata-contents-1 (cdr contents-list) res))
        (else (laml-error "aggregated-ast-cdata-contents: Strange constituent of AST:" (car contents-list)))))



;;; Charater entities.
;;; The function char-ref returns a character-reference. In XML, character references are used to
;;; denote characters which are not directly accessible from normal input devices.

;; Return a character reference value for x. The parameter may be a positive, decimal integer or 
;; a symbolic name (string or symbol), such as amp, gt, or lt. Character references are represented internally
;; as an appropriate list structure, similar to an AST list structure. In xml-render, these structures are
;; converted to  XML's character notation. With this we avoid a conflict with character transformation using
;; the HTML transformation tables.
(define (char-ref x)
 (list 'char-ref
    (cond ((number? x) x)
          ((string? x) (as-symbol x))
          ((symbol? x) x)
          (else (laml-error "char-ref: Invalid parameter:" x)))))


;; Is x a character reference.
(define (char-ref? x)
 (and (list? x) (>= (length x) 2) (eq? (car x) 'char-ref) (or (number? (cadr x)) (symbol? (cadr x)))))

;; Return the number or symbol of char-ref-structure.
(define (char-ref-value char-ref-structure)
  (cadr char-ref-structure))

(define (xml-render-char-ref char-ref)
 (if (not (char-ref? char-ref)) (laml-error "xml-render-char-ref: XML rendering non char-ref:" char-ref))
 (letrec ((char-ref-render 
	   (lambda (x)
	     (cond ((number? x) (string-append "&#" (as-string x) ";")) ; we do not enforce a three-digit representation
		   ((symbol? x) (string-append "&" (as-string x) ";"))
		   (else (laml-error "xml-render-char-ref: the parameter must be numeric, a symbol, or a string" x)))))
          (three-digit-string 
            (lambda (n)
             (cond ((and (>= n 0) (< n 10)) (string-append "00" (as-string n)))
		   ((and (>= n 10) (< n 100)) (string-append "0" (as-string n)))
		   ((< n 1000) (as-string n))
		   (else (error "three-digit-string: parameter must be between 0 and 999")))))
         )
   (char-ref-render (char-ref-value char-ref))))
  

; ---------------------------------------------------------------------------------------------------
;;; XML language bookkeeping.


;; The list of names (symbols) which causes name clashes relative to 
;; the current set of languages in use.
;; Assigned by register-xml-in-laml-language.
;; Thus, this variable is always up-to-date in between registrations of languages.
;; Be careful not to redefine this variable by double loading this file.
(define xml-in-laml-name-clashes '())

;; Register that language (the first parameter) is an XML language in use in the current LAML session.
;; The first parameter, language, is a symbol.
;; The second parameter, language-map, is the language map of the language.
;; As a precondition is a assumed that we do not register the same language more than once.
(define (register-xml-in-laml-language language language-map)
 (set! xml-in-laml-name-clashes (precompute-name-clashes language-map xml-in-laml-name-clashes)) 
 (if (not (memq language (languages-in-use)))
     (set! xml-in-laml-languages-in-use (cons (cons language language-map) xml-in-laml-languages-in-use)))
)

; Return the new name clashes given that new-language-map is about to be registered.
; The existing name clashes is passed as the parameter existing-name-clashes.
(define (precompute-name-clashes new-language-map existing-name-clashes)
  (let* ((all-existing-names (flatten (map element-names-of-language (languages-in-use))))
         (new-names (map car new-language-map))
         (intersection (list-intersection-by-predicate new-names all-existing-names eq?))
        )
    (remove-duplicates-by-predicate (append existing-name-clashes intersection) eq?)))


;; Return the language map of language.
;; If language is not registered via register-xml-in-laml-language, return #f.
;; The parameter language is a symbol.
(define (language-map-of language)
  (defaulted-get language xml-in-laml-languages-in-use #f))

;; is language in use - is it registered as an xml-in-laml language?
(define (language-in-use? language)
  (let ((lg-map (defaulted-get language xml-in-laml-languages-in-use #f)))
    (if lg-map #t #f)))

;; Return the list of language in used, as registered by register-xml-in-laml-language.
(define (languages-in-use) 
  (map car xml-in-laml-languages-in-use))

;; Return the element names of language, as defined by its language map.
(define (element-names-of-language language)
  (let ((lg-map (defaulted-get language xml-in-laml-languages-in-use '())))
    (map car lg-map))) 

;; Is name involed in a name clash among the registered xml-in-laml languages?
(define (causes-xml-in-laml-name-clash? name)
  (memq name xml-in-laml-name-clashes))

(define (activator-via-language-map language)
 (lambda (element-name)
  (let ((lg-map (language-map-of language)))
    (get-mirror-function lg-map element-name))))



; ---------------------------------------------------------------------------------------------------
;;; The language map.
;;; The language map is a mapping that allows us to access the element mirror functions independent of
;;; any name clash. 
;;; In this version the language map is an association list from element names (symbols) to function objects.
;;; In the longer run, it should be more efficient search structure.

; An auxilary variable used by the xml-in-laml tool:
(define temp-language-map '())

; An auxiliry variable used by the xml-in-laml tool:
(define temp-mirror-function #f)

;; Return the mirror function based on element-name from language map.
;; Return #f if no mirror function exists in the map.
;; The parameter element-name is either a symbol or a string.
(define (get-mirror-function language-map element-name)
 (let ((element-name-symbol (as-symbol element-name)))
   (defaulted-get element-name-symbol language-map #f)))

;; Update the language map by the association of element-name and mirror function.
;; Returns the updated map.
;; The parameter element-name is either a symbol or a string.
;; Causes an error if there already is an entry for element-name
(define (put-mirror-function language-map element-name mirror-function)
 (let ((element-name-symbol (as-symbol element-name)))
   (if (get-mirror-function language-map element-name-symbol)
       (laml-error "put-mirror-function: The name" element-name "is defined twice.")
       (cons (cons element-name-symbol mirror-function) language-map))))


; -----------------------------------------------------------------------------------------------------

;;; Mirror generation functions.
;;; The functions in this sections are the important and 'heavy' mirror generation functions.

;; Separates parameters according to the mirror rules of LAML.
;; In other words, this is the function which implements the central LAML mirror function rules.
;; Returns a cons cell of contents and attributes.
;; Attributes are returned in property list format.
;; Contents are returned as a list contents elements (strings, ASTs, booleans).
;; .internal-references "similar function" "xml-sort-superficially-tag-parameters"
;; .form (xml-sort-tag-parameters parameters tag-name [language])
(define (xml-sort-tag-parameters parameters tag-name . optional-parameter-list)
 (let ((language (optional-parameter 1  optional-parameter-list #f)))
  (xml-sort-tag-parameters-1 parameters parameters '() '() tag-name language #f)))

; The superficial? controls the handling of white space suppress. 
; If superficial?, keep white space suppress in asts. If not superficial?, get rid of them.
(define (xml-sort-tag-parameters-1 original-parameters parameters contents attributes tag-name language superficial?)
 (letrec ((strip-initial-explicit-spaces 
	   (lambda (cl) (if (and (pair? cl) (eq? (car cl) explicit-space)) (strip-initial-explicit-spaces (cdr cl)) cl)))
	  (list-non-ast? (lambda (x) (and (list? x) (not (ast? x)))))
         )
  (let ((white-space? (xml-represent-white-space-in? language)))
   (cond ((null? parameters) (cons (reverse (strip-initial-explicit-spaces contents)) (reverse attributes)))


        ; CONTENTS DATA - CDATA, CHAR-REF or AST - and then a space suppress value - no space after
        ((and (contents-data? (car parameters)) (not (null? (cdr parameters)))
              (white-space-related? (cadr parameters)) (white-space-suppress? (cadr parameters)))              
                 (xml-sort-tag-parameters-1 original-parameters (cddr parameters) 
                                        (if white-space? (if superficial? (cons explicit-space-suppress (cons (car parameters) contents)) (cons (car parameters) contents)) (cons (car parameters) contents))
                                        attributes tag-name language superficial?))

        ; CONTENTS DATA - CDATA, CHAR-REF or AST - and then a space value -  space after
        ((and (contents-data? (car parameters)) (not (null? (cdr parameters)))
              (white-space-related? (cadr parameters)) (forced-white-space? (cadr parameters)))                    ; space after
                 (xml-sort-tag-parameters-1 original-parameters
                                        (cddr parameters) (if white-space? (cons explicit-space (cons (car parameters) contents)) (cons (car parameters) contents))
                                        attributes tag-name language superficial?))

        ; CONTENTS DATA - CDATA, CHAR-REF or AST - space after
        ((contents-data? (car parameters))   ; space after
                 (xml-sort-tag-parameters-1 original-parameters 
                                        (cdr parameters) (if white-space? (cons explicit-space (cons (car parameters) contents)) (cons (car parameters) contents))
                                        attributes tag-name language superficial?))


        ; LIST and then a space suppress value - no space after
        ((and (list-non-ast? (car parameters)) (not (null? (cdr parameters))) (white-space-related? (cadr parameters)) (white-space-suppress? (cadr parameters)))
           (let* ((rec-res (xml-sort-tag-parameters-1 original-parameters (car parameters) '() '() tag-name language superficial?))
                  (rec-contents (car rec-res))
		  (rec-attributes (cdr rec-res)) ; reversed
                 )  
              (xml-sort-tag-parameters-1 original-parameters
                                        (cddr parameters) 
                                        (append (if (and white-space? superficial?) (list explicit-space-suppress) '()) (reverse rec-contents) contents)
                                        (append (reverse rec-attributes) attributes) tag-name language superficial?)))

        ; LIST and then a space value - space after
        ((and (list-non-ast? (car parameters)) (not (null? (cdr parameters))) (white-space-related? (cadr parameters)) (forced-white-space? (cadr parameters)))
           (let* ((rec-res (xml-sort-tag-parameters-1 original-parameters (car parameters) '() '() tag-name language superficial?))
                  (rec-contents (car rec-res))
		  (rec-attributes (cdr rec-res)) ; reversed
                 )  
              (xml-sort-tag-parameters-1 original-parameters
                                        (cddr parameters) 
                                        (append (if white-space? (list explicit-space) '()) (reverse rec-contents) contents) 
                                        (append (reverse rec-attributes) attributes) tag-name language superficial?)))

        ; LIST - space after
        ((and (list-non-ast? (car parameters)))
           (let* ((rec-res (xml-sort-tag-parameters-1 original-parameters (car parameters) '() '() tag-name language superficial?))
                  (rec-contents (car rec-res))
		  (rec-attributes (cdr rec-res)) ; reversed
                 ) 
              (xml-sort-tag-parameters-1 original-parameters
                                        (cdr parameters) 
                                        (append (if (and white-space? (not (null? rec-contents))) (list explicit-space) '())
                                                (reverse rec-contents)
                                                contents)
                                        (append (reverse rec-attributes) attributes) tag-name language superficial?)))

        ; SYMBOL - attribute name 
        ((and (symbol? (car parameters)) (not (null? (cdr parameters))) (not (white-space-related? (cadr parameters))))
           (let ((attr-name (car parameters))
                 (attr-val (cadr parameters)))
             (if (and (xml-accept-only-string-valued-attributes-in? language) (not (string? attr-val)))
                 (cond ((ast? attr-val)
                          (xml-check-error (string-append "An element content item of type  " (ast-element-name attr-val) "  is passed as the value of the attribute" (as-string #\newline)
                                                  (as-string attr-name) "  in an instance of the  " tag-name "  element. " (as-string #\newline) "The attribute is ignored.")))
                       (else
                        (xml-check-error (string-append "A non-string value  " (as-string attr-val) "  is passed as the value of the attribute" (as-string #\newline)
                                                  (as-string attr-name) "  in an instance of the  " tag-name "  element." (as-string #\newline) "The attribute value is converted to a string."))))

             )
             (xml-sort-tag-parameters-1 original-parameters (cddr parameters) contents
                                        (if (not (ast? attr-val)) (cons (as-string attr-val) (cons attr-name attributes)) attributes)
                                        tag-name language superficial?)))

        ; Extra superfluous white space or white space suppress value - ignore and just 'eat' it
        ((white-space-related? (car parameters))
             (xml-sort-tag-parameters-1 original-parameters (cdr parameters) contents attributes tag-name language superficial?))

        ; SYMBOL - error
        ((symbol? (car parameters)) (xml-sort-error 
                                      (string-append "XML-in-LAML error: " "Attributes of the  " tag-name
                                      "  element must be of the form  'symbol \"value\" "  (as-string #\newline)
                                    "Only the symbol  " (as-string (car parameters))
                                    "  appears in last encountered attribute.") original-parameters))

        (else (xml-sort-error (string-append "Error in an XML-in-LAML  " tag-name "  element.") original-parameters))
  ))))


;; Like xml-sort-tag-parameters, but collect the content contribution in a relatively raw surface form.
;; Handle attributes in the same way as xml-sort-tag-parameters. As a contrast to xml-sort-tag-parameters, 
;; white space related values are passed unchanged by this funnction.
;; Use this kind of 'sorting' to obtain XML-in-LAML parameter passing in abstractions, which transfer data 
;; to XML-in-LAML mirror functions (which in turn use xml-sort-tag-parameters).
;; .internal-references "similar function" "xml-sort-tag-parameters"
;; .form (xml-sort-superficially-tag-parameters parameters tag-name [language])
(define (xml-sort-superficially-tag-parameters parameters tag-name . optional-parameter-list)
 (let ((language (optional-parameter 1  optional-parameter-list #f)))
  (xml-sort-tag-parameters-1 parameters parameters '() '() tag-name language #t)
))


(define (xml-sort-error message parameters)
 (let* ((max-str-lgt 80)  ; symbolic constant definition
        (parameters-1 (as-quoted-string parameters))
        (parameters-2 (if (> (string-length parameters-1) max-str-lgt) 
                        (string-append (substring parameters-1 0 (- max-str-lgt 1)) "...")
                        parameters-1)))
  (error
    (string-append
      message (as-string #\newline)
      "  The list of parameters: " parameters-2 (as-string #\newline)))))

; -----------------------------------------------------------------------------------------------------------------------

;; Return an XML mirror surface function, in which textual content parameters and attribute
;; value pairs can be given in a very liberal fashion.
;; .parameter validation-procedure the procedure that validates all aspects of the application of the element mirror function.
;; .parameter tag-name a string which represents the name of the tag (used for error message purposes).
;; .parameter default-dtd-attributes an alist of attribute key value pairs, as specified in the DTD. 
;; .parameter single-double-kind either the symbol single, double or possible-single.
;; .parameter language the language in use (a symbol).
;; .parameter overlap-check? controls whether to check for name clashes in between the registered XML-in-LAML languages.
;; .parameter action-procedure serves both as a boolean and a procedure. If non-false, the parameter servers as action procedure, which will receive the AST for action purposes.
(define (generate-xml-mirror-function validation-procedure tag-name default-dtd-attributes single-double-kind language overlap-check? action-procedure)
 (cond ((eq? single-double-kind 'double)
	(lambda parameters
	  (let* ((contents-attributes (xml-sort-tag-parameters parameters tag-name language))
		 (contents (car contents-attributes))
		 (attributes (if (xml-pass-default-dtd-attributes-in? language)
                                 (append (cdr contents-attributes) default-dtd-attributes)
                                 (cdr contents-attributes)))
		)
	    (validation-procedure tag-name attributes contents overlap-check?)
	    (if action-procedure
                (let ((the-ast (make-ast tag-name contents attributes 'double language)))
                  (action-procedure the-ast)
                  the-ast)
                (make-ast tag-name contents attributes 'double language)
            )
	  )
	)
       )

       ((eq? single-double-kind 'single)   
          (lambda parameters
            (let* ((contents-attributes (xml-sort-tag-parameters parameters tag-name language))
		   (contents (car contents-attributes))
		   (attributes (if (xml-pass-default-dtd-attributes-in? language)
                                   (append (cdr contents-attributes) default-dtd-attributes)
                                   (cdr contents-attributes)))
		  )
	      (validation-procedure tag-name attributes contents overlap-check?)
              (if action-procedure
                  (let ((the-ast (make-ast tag-name '() attributes 'single language)))
                    (action-procedure the-ast)
                    the-ast)
	          (make-ast tag-name '() attributes 'single language)
              )
            )
          )
       )

       (else (error (string-append "generate-xml-mirror-function: unknown single-double-kind: " (as-string single-double-kind))))
 )
)



; ---------------------------------------------------------------------------------------------------
;;; Contents validation. 
;;; The definitions in this section carry out the content validation. 


; ; The list on which we accumulate encountered validation problems.
; (define xml-problem-contents-list '())
; 
; ; The last element which caused a problem in xml-member-remember
; (define xml-problem-element #f)
; 
; ; A variant of member which registers context on xml-problem-contents-list in case (not (member el lst))
; (define (xml-member-remember el lst context)
;  (let ((res (member el lst)))
;    (if (not res)
;        (begin
;          (set! xml-problem-contents-list (cons context xml-problem-contents-list))
;          (set! xml-problem-element el)
;        ))
;    res))
; 
; ;; Return a predicate which checks whether the contents matches one or more instances from element-constituent-list.
; ;; Element constituent-list is a list of strings, namely element tag names or
; ;; "#PCDATA".
; ;; A true results means OK - no error
; ;; Avoid a O(n^2) algorithm here in the final implementation
; (define (one-or-more . element-constituent-list)
;   (lambda (contents)
;     (set! xml-problem-contents-list '())
;     (cond ((or (and (boolean? contents) (not contents))
;                (and (list? contents) (null? contents))) 
;              (begin
;                   (set! xml-problem-contents-list (cons "Empty contents not acceptable" xml-problem-contents-list))
;                   #f))  ; empty contents not acceptable
;           ((ast? contents) (xml-member-remember (ast-element-name contents) element-constituent-list contents))
; 	  ((cdata? contents) (xml-member-remember "#PCDATA" element-constituent-list contents))
; 	  ((forced-white-space? contents) #t)
; 	  ((list contents)
;             (and 
; 	     (>= (length contents) 1)	; implement more efficienly
; 	     (accumulate-right
; 	      (lambda (x y) (and x y))
; 	      #t
; 	      (map 
; 	       (lambda (cont)
; 		 (cond ((and (boolean? cont) (not cont))
;                          (begin
; 			   (set! xml-problem-contents-list (cons "Empty contents not acceptable" xml-problem-contents-list))
; 			   #f)) ; even a single #f is not acceptable
;                        ((ast? cont) (xml-member-remember (ast-element-name cont) element-constituent-list cont))
; 		       ((cdata? cont) (xml-member-remember "#PCDATA" element-constituent-list cont))
; 		       ((forced-white-space? cont) #t)
; 		       (else #f)))
; 	       contents)))
;           )
; 	  (else (laml-error "one-or-more: Unknown contents: " (as-string contents))))))
; 
; ;; Return a predicate which checks whether the contents matches zero or more instances from element-constituent-list
; (define (zero-or-more . element-constituent-list)
;   (lambda (contents)
;     (set! xml-problem-contents-list '())
;     (cond ((and (boolean? contents) (not contents)) #t)  ; no contents, which is acceptable
;           ((ast? contents) (xml-member-remember (ast-element-name contents) element-constituent-list contents))
; 	  ((cdata? contents) (xml-member-remember "#PCDATA" element-constituent-list contents))
; 	  ((forced-white-space? contents) #t)
; 	  ((list contents)
; 	    (accumulate-right
; 	     (lambda (x y) (and x y))
; 	     #t
; 	     (map 
; 	      (lambda (cont)
; 		(cond ((and (boolean? cont) (not cont)) #t) ; list no contents, which is acceptable
;                       ((ast? cont) (xml-member-remember (ast-element-name cont) element-constituent-list cont))
; 		      ((cdata? cont) (xml-member-remember "#PCDATA" element-constituent-list contents))
; 		      ((forced-white-space? cont) #t)
; 		      (else #f)))
; 	      contents))
; 	   )
; 	  (else (laml-error "one-or-more: Unknown contents: " (as-string contents))))))
; 
; 
; 
; ;; Return a predicate which checks whether the contents match the sequential rhs-elements.
; ;; rhs-elements is a list of strings or symbols, where symbols represent optionals.
; ;; Example (sequence-with-optionals "xxx" "yyy" 'zzz) corresponds to
; ;;         xxx, yyy, zzz?
; ;; We assume that rhs-elements, in the starting point, always contains at least one element.
; (define (sequence-with-optionals . rhs-elements)
;   (lambda (contents)
;     (set! xml-problem-contents-list '())
;     (let ((all-optionals? (lambda (rhs-elements) (accumulate-right and-fn #t (map symbol? rhs-elements)))))
;      (cond ((or (and (boolean? contents) (not contents))
; 		(and (list? contents) (null? contents) (not (all-optionals? rhs-elements)))) 
; 	    (xml-add-problem! (xml-enrich-error-message "Empty contents not acceptable in sequence" contents))
; 	    ) 
; 	   ((cdata? contents) (xml-add-problem! (xml-enrich-error-message "PCDATA not acceptable in this context" contents)))
; 	   ((forced-white-space? contents) (xml-add-problem! "Empty contents not acceptable in sequence"))
; 	   ((list contents)
;             (and 
; 	     (>= (length contents) 1)	
;              (sequence-with-optionals-1 contents rhs-elements)
; 	     )
; 	    )
; 	   (else (laml-error "sequence-with-optionals: Unknown contents: " (as-string contents)))))))
; 
; ; A tail recursive helping function to sequence-with-optionals.
; ; contents-list is a list contents elements.
; ; rhs-elements is a list of sequential elements of the parsed rhs (strings or symbol lists).
; (define (sequence-with-optionals-1 contents-list rhs-elements)
;   (cond ((and (null? contents-list) (null? rhs-elements)) #t) ; OK
;         ((and (null? contents-list) (remaining-optionals? rhs-elements)) #t) ; content-list end - one optional rhs el. OK.
;         ((null? rhs-elements) (xml-add-problem! (xml-enrich-error-message "Sequence too long" contents-list)))
;         ((null? contents-list) (xml-add-problem! (xml-enrich-error-message "Sequence too short. Missing" rhs-elements)))
; 
;         ; now both lists non-empty
; 
;         ((cdata? (car contents-list))
;             (xml-add-problem! (xml-enrich-error-message "Encountered PCDATA where sequence is expected" (string-it (car contents-list)))))
; 
;         ((forced-white-space? (car contents-list)) (sequence-with-optionals-1 (cdr contents-list) rhs-elements))
; 
;         ((and (ast? (car contents-list)) (string? (car rhs-elements)) 
;               (equal? (ast-element-name  (car contents-list)) (car rhs-elements)))  ; two matching elements - proceed booth
;            (sequence-with-optionals-1 (cdr contents-list) (cdr rhs-elements)))
; 
;         ((and (forced-white-space? (car contents-list)))  ; forced white space - proceed
;            (sequence-with-optionals-1 (cdr contents-list) rhs-elements))
; 
;         ((and (ast? (car contents-list)) (symbol? (car rhs-elements)) 
;               (equal? (ast-element-name  (car contents-list)) (as-string (car rhs-elements)))) ; match against optional - proceed booth
;            (sequence-with-optionals-1 (cdr contents-list) (cdr rhs-elements)))
; 
;         ((and (ast? (car contents-list)) (symbol? (car rhs-elements)) 
;               (not (equal? (ast-element-name  (car contents-list)) (as-string (car rhs-elements))))) ; non match against optional - proceed rhs only
;            (sequence-with-optionals-1 contents-list (cdr rhs-elements)))
; 
;         (else  ; all other cases cause mismatch
;            (xml-add-problem! (xml-enrich-error-message "Mismatch in sequence" contents-list) ))))
; 
; (define (remaining-optionals? lst)
;   (cond ((null? lst) #t)
;         (else (and (symbol? (car lst)) (remaining-optionals? (cdr lst))))
;   )
; )
; 
; ;; A predicate which checks whether the contents is PCDATA.
; (define (pcdata-checker contents)
;  (set! xml-problem-contents-list '())
;  (let ((contents-1 (if (list? contents) (filter (negate white-space-related?) contents) contents)))
;   (cond ((and (list? contents) (null? contents))
;           (begin
; 	   (set! xml-problem-contents-list 
; 		 (cons (xml-enrich-error-message "The instance should not be empty" contents-1)
; 		       xml-problem-contents-list))
; 	   #f))
;         ((cdata? contents-1)		; #f - no contents probably
;            #t)
;         ((not (list? contents-1))
;           (begin
; 	   (set! xml-problem-contents-list 
; 		 (cons (xml-enrich-error-message "The instance should not be empty" contents-1)
; 		       xml-problem-contents-list))
; 	   #f))
; 	((list? contents-1)
; 	 (if (not
; 	       (accumulate-right and-fn #t (map cdata? contents-1))) ; all cdata?
; 	     (begin
; 	       (set! xml-problem-contents-list 
; 		     (cons (xml-enrich-error-message 
;                              "The instance must consist of textual contents - no embedded elements allowed" contents-1)
; 			   xml-problem-contents-list))
; 	       #f)
; 	     #t))
; 	(else #f))))
; 
; 
; ;; The procedure for reporting validation problems.
; ;; Adds error message to xml-problem-contents-list and returns #f, as an indication of validation failure.
; (define (xml-add-problem! error-message)
;  (set! xml-problem-contents-list (cons error-message xml-problem-contents-list))
;  #f)
; 
; ;; Enrich an error message with info about a constitutent (AST)
; (define (xml-enrich-error-message error-message constituent)
;   (string-append error-message ":   "  (xml-render-1 constituent)))
; 
; ;; Enrich a list of error messages with info about a constitutent (AST)
; (define (xml-enrich-error-messages error-message-list constituent)
;   (if (list? error-message-list)
;       (map (lambda (e) (string-append e ":   " (xml-render-1 constituent))) error-message-list)
;       error-message-list))
; 
; 
; 
; ; Is there an AST with el-name in the list of ASTs.
; ; precondition: contents is a list of ASTs
; (define (there-exists-ast? el-name contents-list)
;   (find-in-list (lambda (ast) (equal? (ast-element-name ast) el-name)) contents-list))
; 
; ; Return the instances of ast element name in the contents list.
; ; precondition: contents is a list of ASTs
; (define (instances-of-ast el-name contents-list)
;   (filter (lambda (ast) (equal? (ast-element-name ast) el-name)) contents-list))
; 
; 
; ; Do all ASTs have element names in el-name-list
; ; precondition: contents is a list of ASTs
; ; Should not be O(n^2) in a more finished implementation
; (define (all-ast-member-of? el-name-list contents-list)
;  (accumulate-right and-fn #t 
;   (map (lambda (ast) (member (ast-element-name ast) el-name-list)) contents-list)))
; 
; ; A version of filter which work normally on lists, and else just returns the empty list.
; (define (filter-contents pred contents-list)
;  (if (list? contents-list)
;      (filter pred contents-list)
;      '()))
; 
; 
; ; Apply validation-predicate on contents. If the validation predicate returns a false result
; ; issue a warning.
; (define (validate-contents! contents validation-predicate tag-name)
;  (set! xml-problem-contents-list '())
;  (let ((validation-status (validation-predicate contents)))
;   (if (not validation-status)
;       (let ((count (length xml-problem-contents-list)))
;         (if (> count 0) ; should always be true
;           (xml-check-error
;            (string-append 
;             (string-append 
;               (if (= count 1) 
; 		  "XML validation error(s) encountered in an instance of the  "
; 		  (string-append  "XML validation error(s) encountered in an instance of the  "))
;               (as-string tag-name) "  element")
;             (indented-terminal-lines (map (compose truncate-string xml-render-1) xml-problem-contents-list)))))))))


; ---------------------------------------------------------------------------------------------------------------
; XML document validation.
; The non-trivial validation cases are based on the automata generated by tools/xml-in-laml/xml-in-laml.scm.


; The symbol which is assumed to end all input to the dfa. 
; The value of this variable must comply with the similar constant in tools/xml-in-laml/xml-in-laml.scm.
(define terminator-symbol 'terminator$$)

; The symbol that represents textual contents (pcdata) during validation.
(define textual-content-symbol 'textual-contents$$)


;; Apply the deterministic final state automaton dfa on contents. 
;; If contents is not accepted call xml-check-error for an appropriate reporting of the validation error.
;; .parameter contents A list of content items, such as strings, ASTs and white space markers.
;; .parameter dfa A deterministic final state automation which controls the acceptance.
;; .parameter tag-name The name of the element - used for error message purposes. 
(define (validate-contents-by-dfa! contents dfa tag-name)
 (if (not (list? contents)) (laml-error "Contents passed to validate-contents-by-dfa! is assumed to be a list"))
 (let* ((augmented-contents (xml-prepare-contents-for-validation contents))
        (accepted (automaton-accepts? dfa augmented-contents)))
  (if (not accepted)
      (cond ((not last-automaton-input-symbol)
               (xml-check-error "Empty and insufficient input to" (as-string tag-name) "construct:" (as-string #\newline)
                        "  " (truncate-string (xml-render-1 contents))))
            ((eq? last-automaton-input-symbol terminator-symbol)
               (xml-check-error "Abrupt termination of" (a-or-an (as-string tag-name)) (as-string tag-name) "construct:" (as-string #\newline)
                        "  " (truncate-string (xml-render-1 contents))))
            ((eq? last-automaton-input-symbol textual-content-symbol)
               (let ((the-textual-contents (list-ref (filter (negate white-space-related?) contents) (max 0 (- automaton-input-number 1)))))
                (xml-check-error "Textual contents" (string-it (xml-render-1 the-textual-contents)) "is illegal in" (a-or-an (as-string tag-name)) 
                                  (as-string tag-name) "construct:" (as-string #\newline)
                        "  " (truncate-string (xml-render-1 contents)))))
            (else 
               (xml-check-error "Encountered a misplaced" (as-string last-automaton-input-symbol) "in" (a-or-an (as-string tag-name)) (as-string tag-name) "construct:" (as-string #\newline)
                        "  " (truncate-string (xml-render-1 contents))))))))


;; Validate that contents is pure PCDATA.
;; .parameter contents A list of content items, such as strings, ASTs and white space markers.
;; .parameter tag-name The name of the element - used for error message purposes. 
(define (validate-as-pcdata! contents tag-name)
  (if (not (list? contents)) (laml-error "Contents passed to validate-as-pcdata! is assumed to be a list"))
  (let ((res (do-validate-pcdata-contents contents)))
    (cond ((or (symbol? res) (string? res))
             (xml-check-error "Encountered a misplaced " (as-string res) " in " (a-or-an (as-string tag-name))
                              (as-string tag-name) " construct, where only textual contents is allowed." (as-string #\newline)
                "  "  (truncate-string (xml-render-1 contents))))
          ((and (boolean? res) (not res))
             (xml-check-error "Unindentified problem in " (as-string tag-name) (as-string #\newline)
                "  "  (truncate-string (xml-render-1 contents)))))))

(define (do-validate-pcdata-contents contents)
  (call-with-current-continuation 
    (lambda (exit)
     (do-validate-pcdata-contents-1 contents exit))))

(define (do-validate-pcdata-contents-1 contents return)
  (cond ((null? contents) #t) 
        (else (let ((content-item (car contents)))
                 (cond ((ast? content-item)  (return (ast-element-name content-item)))
       	               ((cdata? content-item) #t)
                       ((char-ref? content-item) #t)
		       ((white-space-related? content-item) #t)
		       (else (return #f)))
                 (do-validate-pcdata-contents-1 (cdr contents) return)))))

;; Validate that contents corresponds to mixed contents. Mixed contents includes PCDATA a number of other choices in a 'zero-or-more' structure.
;; It means that it is PCDATA or one of a number of possible ASTs.
;; .parameter contents A list of content items, such as strings, ASTs and white space markers.
;; .parameter tag-name The name of the element - used for error message purposes. 
;; .parameter symbol-choice-list A list of symbols for the possible choices - a list of symbols.
(define (validate-mixed-contents-by-simple-means! contents symbol-choice-list tag-name)
  (if (not (list? contents)) (laml-error "Contents passed to validate-mixed-contents-by-simple-means! is assumed to be a list"))
  (let ((res (do-validate-mixed-contents contents symbol-choice-list)))
    (cond ((or (symbol? res) (string? res))
             (xml-check-error "Encountered a misplaced" (as-string res) "in" (as-string tag-name) "construct." (as-string #\newline) 
                "  "  (truncate-string (xml-render-1 contents))))
          ((and (boolean? res) (not res))
             (xml-check-error "Unindentified problem in" (as-string tag-name) (as-string #\newline)
                "  "  (truncate-string (xml-render-1 contents)))))))
   

(define (do-validate-mixed-contents contents symbol-choice-list)
  (call-with-current-continuation 
    (lambda (exit)
     (do-validate-mixed-contents-1 contents symbol-choice-list exit))))

(define (do-validate-mixed-contents-1 contents symbol-choice-list return)
  (cond ((null? contents) #t) 
        (else (let ((content-item (car contents)))
                 (cond ((ast? content-item) (if (memq (as-symbol (ast-element-name content-item)) symbol-choice-list) #t (return (ast-element-name content-item))))
       	               ((cdata? content-item) #t)
                       ((char-ref? content-item) #t)
		       ((white-space-related? content-item) #t)
		       (else (return #f)))
                 (do-validate-mixed-contents-1 (cdr contents) symbol-choice-list return)))))


(define (xml-prepare-contents-for-validation contents)
  (append
    (map (lambda (content-item)
           (cond ((ast? content-item) (as-symbol (ast-element-name content-item)))
                 ((cdata? content-item) textual-content-symbol)
                 ((char-ref? content-item) textual-content-symbol)
                 (else (laml-error "xml-prepare-contents-for-validation: Unknown element content item:" content-item))))
         (filter (negate white-space-related?) contents))
    (list terminator-symbol)))

; Empty contents checking of empty elements
(define (xml-check-for-empty-contents! contents tag-name)
  (if (not (null? contents))
      (xml-check-error 
        (string-append "The empty element " tag-name " is not supposed to have any contents: "
                       (xml-render-1 contents) ". The element content is ignored."))))

; Return "a" or "an" depending on following-word
(define (a-or-an following-word)
  (cond ((blank-string? following-word) "a")
        ((> (string-length following-word) 0)
         (let ((first-char (string-ref following-word 0)))
           (if (memq first-char (list #\a #\e #\i #\o #\u #\y ))
               "an"
               "a")))
        (else "a")))


(define (indented-terminal-lines line-list)
 (let ((sep (string-append (as-string #\newline) "  ")))
  (string-append sep (list-to-string line-list sep))))

; A slight generalization of xml-render
(define (xml-render-1 contents)
  (cond ((ast? contents) (xml-render contents))
	((cdata? contents) contents)
        ((char-ref? contents) (xml-render-char-ref contents))
	((forced-white-space? contents) "SPACE")
	((list? contents) (list-to-string (map xml-render-1 contents) " "))
        (else "??")))

(define (truncate-string str)
 (if (> (string-length str) xml-error-truncation-length)
     (string-append (substring str 0 xml-error-truncation-length) "...")
     str))

; ; A predicate which always gives positive validation.
; ; Used if the xml-in-laml tool constant use-manually-programmed-validation-predicates is #f.
; (define (faked-generice-content-checker x)
;   #t)


; and as a function - as opposed to a macro.
(define (and-fn x y) (and x y))


; --------------------------------------------------------------------------------------------------------------------------------
;;; Attribute checking. 
;;; The definitions in this section is used by the validation procedures, which are specific
;;; for each supported XML language.
;;; <em> Current status of attribute checking</em>: TokenizedTypes are not checked (ID, IDREFS, ...). In reality there
;;; is probably not much to check (besides some lexical checks which we can easily deal with, and some contraints which
;;; are much harder). NotationTypes are not dealt with - mostly because I have never encountered them, so the motivation to
;;; program check for this special enumeration type is not high. The #FIXED default keyword is present in the parsed DTD
;;; information, but we do not carry out any check against it (which is: If the attribute is given it must have the fixed default value).
;;; The attribute value normalization called for in section 3.3.3 of the XML specificaiton is not done either. Attribute duplication
;;; is checked for, as well as presence of angle characters in attribute values.

; Attribute tripple selectors
(define att-name (make-selector-function 1))
(define att-type (make-selector-function 2))
(define att-status (make-selector-function 3))

;; Check the attributes (first par) in the calling form against the attribute definitions taken from the DTD file (second par).
;; The first parameter, attribute, is the attributes of the calling form, on property list form:
;; (a1 v1 ... an vn). ai is a symbol and vi is a string.
;; The second parameter dtd-attribute-definition is the attributes as defined
;; in the dtd for the element in question. The third parameter number-of-req-attributes is the number
;; of required attributes in dtd-attr-definitions. It happens to be the case that all the required attributes
;; are located in the front of dtd-attribute-definition.
;; The fourth parameter, tag-name, is the name of the enclosing tag. dtd-attribute-definition is a list of 
;; triples (attr-name attr-type require-status). attr-name is a string, attr-type is
;; a string or a list of strings (possibilities), and require-status is a string
;; such as "#IMPLIED" or "#REQUIRED"
(define (xml-check-attributes! attributes dtd-attribute-definition number-of-req-attributes tag-name)
 (if (even? (length attributes))  ; if not, we report the error elsewhere
  (let ((required-attribute-names 
           (map (compose as-symbol att-name) (front-sublist dtd-attribute-definition number-of-req-attributes)))
        (dtd-attribute-names (map (compose as-symbol car) dtd-attribute-definition))
        (attribute-names (if (null? attributes) '() (every-second-element attributes)))
        (attribute-values (if (null? attributes) '() (every-second-element (cdr attributes))))
       )
    (xml-check-required-attributes! attribute-names required-attribute-names tag-name)
    (xml-check-for-attribute-existence! attribute-names dtd-attribute-names tag-name)
    (xml-check-for-attribute-types! attribute-names attribute-values dtd-attribute-definition tag-name)
    (xml-check-for-attribute-duplicates! attribute-names tag-name)
)))


; Check that all elements in required-attribute-names are in fact present in attribute-names.
; The last parameter, tag-name is the name of the enclosing tag - used for error message purposes.
(define (xml-check-required-attributes! attribute-names required-attribute-names tag-name)
  (if (not (null? required-attribute-names))
      (begin
         (xml-check-one-required-attribute! attribute-names (car required-attribute-names) tag-name)
         (xml-check-required-attributes! attribute-names (cdr required-attribute-names) tag-name))))

(define (xml-check-one-required-attribute! attribute-names required-attribute tag-name)
  (if (not (memq required-attribute attribute-names))
      (xml-check-error 
        (string-append "The required attribute  " (as-string required-attribute) "  is not present in the " tag-name " element."))))

;;; XML warning procedures.

;; A variant of display-warning which prefixes the warning text with a 'XML Warning' prefix.
(define (display-xml-warning . messages)
 (display (string-append "XML Warning: " (laml-aggregate-messages messages))) (newline))


;; The procedure that reports XML validation errors.
;; The default is display-xml-warning, which issues non-fatal warning messages.
;; As an alternative you can use laml-error which provides for fatal error messaging.
;; Both of these accept an arbitrary number of parameters, which are string converted prior to string appending them.
;; .form (xml-check-error . messages)
;; .internal-references "default value" "display-xml-warning"
;; .reference "alternative value" "laml-error" "../../man/general.html#laml-error"
(define xml-check-error display-xml-warning)


; Check that all names in attribute-names found in the list of names dtd-attribute-names
; tag-name is the name of the enclosing tag
; CSS attributes should not be checked for existence. 
(define (xml-check-for-attribute-existence! attribute-names dtd-attribute-names tag-name)
   (if (not (null? attribute-names))
       (begin
         (xml-check-one-attribute-existence! (car attribute-names) dtd-attribute-names tag-name)
         (xml-check-for-attribute-existence! (cdr attribute-names) dtd-attribute-names tag-name))))

(define (xml-check-one-attribute-existence! name dtd-attribute-names tag-name)
  (if (and (not (xml-css-key? name)) (not (memq name dtd-attribute-names)))
      (xml-check-error (string-append "The XML attribute  " (as-string name) "  is not valid in the "  tag-name " element."))))

; Check that the attributes (as splitted into attribute-names and attribute-values - lists of equal lengths)
; have correct types accoring the dtd-attribute-names.
(define (xml-check-for-attribute-types! attribute-names attribute-values dtd-attributes tag-name)
  (if (not (null? attribute-names))
      (begin
         (xml-check-one-attribute-type! (car attribute-names) (car attribute-values) dtd-attributes tag-name)
         (xml-check-for-attribute-types! (cdr attribute-names) (cdr attribute-values) dtd-attributes tag-name))))

(define (xml-check-one-attribute-type! name value dtd-attributes tag-name)
  (let ((attribute-descriptor 
          (find-in-list (lambda (tripple) (eq? (as-symbol (att-name tripple)) name)) dtd-attributes)))
    (if attribute-descriptor
        (xml-check-attribute-value! name value (att-type attribute-descriptor) tag-name))))

; Is value a 'member of' attribute-type. value is a string.
; Is value free of '<' characters.
(define (xml-check-attribute-value! name value attribute-type tag-name)

  ; First check that '<' does not appear in the attribute value:
  (if (not (string-of-negative-char-list? value (list #\<))) ; '<' occurs in the attribute value
      (xml-check-error 
         (string-append "The value  " (string-it value) "  of the XML attribute  "
                        (as-string name) "  is invalid in the  "  tag-name "  element due to an occurrence of '<'.")))

  ; Then check the attribute type:
  (cond ((list? attribute-type) 
            (if (not (member value attribute-type))  
                (xml-check-error (string-append "The value  " (string-it value) "  of the XML attribute  "
                                            (as-string name) "  is not valid in the "  tag-name " element."))))

        ((equal? "CDATA" attribute-type)
            #t)

        ((member attribute-type (list "ID" "IDREF" "IDREFS" "ENTITY" "ENTITIES" "NMTOKEN" "NMTOKENS"))
            #t)

        (else (xml-check-error (string-append "DTD error!!! The type  " (string-it attribute-type) "  of the XML attribute  "
                                            (as-string name) "  is not valid in the "  tag-name " element.")))
  ))

(define (xml-check-for-attribute-duplicates! attribute-names tag-name)
 (let ((duplicated-attribute-names (duplicates-by-predicate attribute-names eq?)))
   (cond ((and (not (null? duplicated-attribute-names)) (null? (cdr duplicated-attribute-names))) ; only one duplicated attribute
           (xml-check-error 
	    (string-append "The attribute  " (as-string (car duplicated-attribute-names))
                           "  is not allowed to appear more than once in a  " tag-name "  element.")))
         ((not (null? duplicated-attribute-names))  ; two or more duplicated attribut names
           (xml-check-error 
	    (string-append "The attributes  " (list-to-string (map as-string duplicated-attribute-names) ", ")
                           "  are not allowed to appear more than once in a  " tag-name "  element.")))
         (else #t) ; do nothing
   )))




; ---------------------------------------------------------------------------------------------------------------
; Check of language overlap

(define (check-language-overlap! name)
  (if (causes-xml-in-laml-name-clash? name)
      (xml-check-error 
       (string-append "The mirror function named " (as-string name) " is ambiguous. Please use it via an appropriate language map."))))


; ---------------------------------------------------------------------------------------------------------------
; Html character transformation table mutations.
; Mutations that cause literal presentation of the characters #\< , #\>, #\", #\' and #\&

(set-html-char-transformation-entry! html-char-transformation-table (char->integer #\<) "&lt;")
(set-html-char-transformation-entry! html-char-transformation-table (char->integer #\>) "&gt;")
(set-html-char-transformation-entry! html-char-transformation-table (char->integer #\") "&quot;")
(set-html-char-transformation-entry! html-char-transformation-table (char->integer #\') "&#39;")
(set-html-char-transformation-entry! html-char-transformation-table (char->integer #\&) "&amp;")

; ---------------------------------------------------------------------------------------------------------------

;;; Rendering functions and procedures.
;;; The functions and procedures in this section transform the internal document representation
;;; (abstract syntax trees) to a textual form (text strings).

; ---------------------------------------------------------------------------------------------------
; Fast AST rendering.

;; Render the XML clause (an AST) to output-port.
;; output-port is assumed to be open. output-port will not be closed by this procedure.
;; .form (render-to-output xml-clause output-port [prolog epilog])
;; .parameter xml-clause an AST
;; .parameter output-port an open output port
;; .parameter prolog The symbol prolog, in which case (standard-prolog) is inserted, or a prolog string to be inserted
;; .parameter epilog The symbol epilog in which case (standard-epilog) is inserted, or a epilog string to be inserted
;; .reference "standard-prolog" "laml.scm" "../../../man/laml.html#standard-prolog"
;; .reference "standard-epilog" "laml.scm" "../../../man/laml.html#standard-epilog"
(define (render-to-output-port xml-clause output-port . optional-parameter-list)
 (let* ((prolog (optional-parameter 1 optional-parameter-list #f))
        (epilog (optional-parameter 2 optional-parameter-list #f))
        (language (if (ast? xml-clause) (ast-language xml-clause) #f))  ; if - just to be conservative.
        (prolog-text (cond ((and (symbol? prolog) (eq? prolog 'prolog)) (standard-prolog language))
                           ((string? prolog) prolog)
                           (else "")))
        (epilog-text (cond ((and (symbol? epilog) (eq? epilog 'epilog)) (standard-epilog))
                           ((string? epilog) epilog)
                           (else "")))
        (put-fn (put-in-sink-stream-generator output-port))
       )
  (put-fn prolog-text)
  (render-fast xml-clause put-fn #f)
  (put-fn epilog-text)))

;; Pretty print the XML clause (an AST) to output-port.
;; output-port is assumed to be open. output-port will not be closed by this procedure.
;; The constants preferred-maximum-width and indentation-delta affect the pretty printing.
;; .form (pretty-render-to-output-port xml-clause output-port [prolog epilog])
;; .parameter xml-clause an AST
;; .parameter output-port an open output port
;; .parameter prolog The symbol prolog, in which case (standard-prolog) is inserted, or a prolog string to be inserted
;; .parameter epilog The symbol epilog in which case (standard-epilog) is inserted, or a epilog string to be inserted
;; .reference "standard-prolog" "laml.scm" "../../../man/laml.html#standard-prolog"
;; .reference "standard-epilog" "laml.scm" "../../../man/laml.html#standard-epilog"
(define (pretty-render-to-output-port xml-clause output-port . optional-parameter-list)
 (let* ((prolog (optional-parameter 1 optional-parameter-list #f))
        (epilog (optional-parameter 2 optional-parameter-list #f))
        (language (if (ast? xml-clause) (ast-language xml-clause) #f))  ; if - just to be conservative.
        (prolog-text (cond ((and (symbol? prolog) (eq? prolog 'prolog)) (standard-prolog language))
                           ((string? prolog) prolog)
                           (else "")))
        (epilog-text (cond ((and (symbol? epilog) (eq? epilog 'epilog)) (standard-epilog))
                           ((string? epilog) epilog)
                           (else "")))
        (put-fn (put-in-sink-stream-generator output-port))
       )
  (put-fn prolog-text)
  (pp-render-fast xml-clause put-fn #f 0 #f)
  (put-fn epilog-text)))

;; Render the xml-clause (an LAML AST) and return the rendered string. In this context, rendering means
;; linearization of the AST to its textual form.
;; .form (xml-render xml-clause [prolog epilog])
;; .parameter xml-clause an AST
;; .parameter prolog The symbol prolog, in which case (standard-prolog) is inserted, or a prolog string to be inserted
;; .parameter epilog The symbol epilog in which case (standard-epilog) is inserted, or a epilog string to be inserted
(define (xml-render xml-clause . optional-parameter-list)
 (let* ((prolog (optional-parameter 1 optional-parameter-list #f))
        (epilog (optional-parameter 2 optional-parameter-list #f))
        (prolog-text (cond ((and (symbol? prolog) (eq? prolog 'prolog)) (standard-prolog))
                           ((string? prolog) prolog)
                           (else "")))
        (epilog-text (cond ((and (symbol? epilog) (eq? epilog 'epilog)) (standard-epilog))
                           ((string? epilog) epilog)
                           (else "")))
       )
   (reset-sink-string)
   (render-fast xml-clause put-in-sink-text-string #f)
   (string-append
     prolog-text
     (sink-string)  
     epilog-text)))



;; Pretty print xml-clause (an LAML AST) and return the rendered string. In this context, rendering means
;; linearization of the AST to its textual, pretty printed form.
;; The constants preferred-maximum-width and indentation-delta affect the pretty printing.
;; .form (pretty-xml-render xml-clause [prolog epilog])
;; .parameter xml-clause an AST
;; .parameter prolog The symbol prolog, in which case (standard-prolog) is inserted, or a prolog string to be inserted
;; .parameter epilog The symbol epilog in which case (standard-epilog) is inserted, or a epilog string to be inserted
(define (pretty-xml-render xml-clause . optional-parameter-list)
 (let* ((prolog (optional-parameter 1 optional-parameter-list #f))
        (epilog (optional-parameter 2 optional-parameter-list #f))
        (prolog-text (cond ((and (symbol? prolog) (eq? prolog 'prolog)) (standard-prolog))
                           ((string? prolog) prolog)
                           (else "")))
        (epilog-text (cond ((and (symbol? epilog) (eq? epilog 'epilog)) (standard-epilog))
                           ((string? epilog) epilog)
                           (else "")))
       )
   (reset-sink-string)
   (pp-render-fast xml-clause put-in-sink-text-string #f 0 #f)
   (string-append
     prolog-text
     (sink-string)  
     epilog-text)))



; --------------------------------------------------------------------
; String sink in terms of segments (a list) of strings.

(define sink-string-segment-size 20000)
(define sink-string-segment-limit (- sink-string-segment-size 1))
(define output-sink-segments '())
(define output-sink-string (make-string sink-string-segment-size))
(define next-sink-ptr 0)

; Put x (string or char) into output sink.
(define (put-in-sink-text-string x)
  (cond ((char? x) 
          (begin
	   (string-set! output-sink-string next-sink-ptr x)
	   (set! next-sink-ptr (+ 1 next-sink-ptr))
           (if (= next-sink-ptr sink-string-segment-size) (prepare-next-sink-segment))))

        ((string? x)
          (do ((lgt (string-length x))
	       (i 0 (+ i 1))
               (j next-sink-ptr (if (< j sink-string-segment-limit) (+ j 1) 0))
              )
	      ((= i lgt) (set! next-sink-ptr j))
	       (string-set! output-sink-string j (string-ref x i))
               (if (= j sink-string-segment-limit) (prepare-next-sink-segment))
              ))

        (else "put-in-sink-string: Can only output chars or strings")))

(define (prepare-next-sink-segment)
 (set! output-sink-segments (cons (string-copy output-sink-string) output-sink-segments))
 (set! next-sink-ptr 0)
)

; Return the resulting sink string
(define (sink-string)
 (string-append 
   (list-to-string (reverse output-sink-segments) "")
   (substring output-sink-string 0 next-sink-ptr)))

(define (reset-sink-string)
 (set! output-sink-segments '())
 (set! next-sink-ptr 0))

; --------------------------------------------------------------------
; Stream sink.

; Generate an output port put procedure
(define (put-in-sink-stream-generator port)
 (lambda (x)
  (cond ((char? x) 
          (write-char x port))
        ((string? x)
          (do ((lgt (string-length x))
               (i 0 (+ i 1))
              )
	      ((= i lgt) 'done)
  	       (write-char (string-ref x i) port)))
        (else "put-in-sink-stream: Can only output chars or strings"))))


; --------------------------------------------------------------------
; Sink independet raw rendering.

(define quote-char #\")

(define (render-fast ast put always-render-white-space?)
 (let* ((tag-name (ast-element-name ast)) 
        (contents-list (ast-subtrees ast))
        (attribute-properlist (ast-attributes ast))  ; split in normal and css here
        (attribute-alist (propertylist-to-alist attribute-properlist))
        (reorganized-attribute-alist (html-css-split attribute-alist '() '()))
        (html-attribute-alist (car reorganized-attribute-alist))
        (css-attribute-alist  (cdr reorganized-attribute-alist))
        (kind (ast-kind ast))

        (language (ast-language ast))
        (xml-transliterate-character-data? (xml-transliterate-character-data-in? language))
        (xml-non-transliteration-elements (xml-non-transliteration-elements-in language))
        (xml-preformatted-text-elements (xml-preformatted-text-elements-in language))  ; (list "pre")
        (xml-char-transformation-table (xml-char-transformation-table-in language))

       )
  (cond ((eq? kind 'single)  ; empty element
          (put #\<) 
          (put tag-name) 

          ; Linearize HTML attributes - duplicate code inlined for double elements!
          (for-each 
            (lambda (attr-pair) 
             (let ((key (car attr-pair))
		   (val (cdr attr-pair)))
               (put #\space)
	       (put (symbol->string key))
	       (put #\=) (put quote-char)
	       (do ((lgt (string-length val))
		    (i 0 (+ i 1))
		   )
		   ((= i lgt) 'done)
		 (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
	       (put quote-char))
            )
          html-attribute-alist)

          ; Linearize CSS attributes - duplicate code inlined for double elements!
          (if (not (null? css-attribute-alist))
	      (begin
                (put #\space) 
		(put "style=") (put quote-char)
		(for-each
                 (lambda (attr-pair) 
                   (let* ((key (symbol->string (car attr-pair)))
			  (non-cssed-key (substring key 4 (string-length key)))
  			  (val (cdr attr-pair)))
		     (put non-cssed-key)
		     (put #\:)
		     (do ((lgt (string-length val))
			  (i 0 (+ i 1))
			 )
			 ((= i lgt) 'done)
		       (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
		     (put #\;))) 
                css-attribute-alist)
		(put quote-char)))

          (put "/>"))

        ((eq? kind 'double)  ; non-empty element
          (put #\<) 
          (put tag-name) 

          ; Linearize HTML attributes:
          (for-each 
            (lambda (attr-pair) 
             (let ((key (car attr-pair))
		   (val (cdr attr-pair)))
               (put #\space)
	       (put (symbol->string key))
	       (put #\=) (put quote-char)
	       (do ((lgt (string-length val))
		    (i 0 (+ i 1))
		   )
		   ((= i lgt) 'done)
		 (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
	       (put quote-char))
            )
          html-attribute-alist)

          ; Linearize CSS attributes:
          (if (not (null? css-attribute-alist))
	      (begin
                (put #\space) 
		(put "style=") (put quote-char)
		(for-each
                 (lambda (attr-pair) 
                   (let* ((key (symbol->string (car attr-pair)))
			  (non-cssed-key (substring key 4 (string-length key)))
  			  (val (cdr attr-pair)))
		     (put non-cssed-key)
		     (put #\:)
		     (do ((lgt (string-length val))
			  (i 0 (+ i 1))
			 )
			 ((= i lgt) 'done)
		       (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
		     (put #\;))) 
                css-attribute-alist)
		(put quote-char)))

          (put #\>)
          (linearize-contents-list-fast contents-list put tag-name 
               xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table (or always-render-white-space? (member tag-name xml-preformatted-text-elements)))
          (put "</")  (put tag-name) (put #\>))

        (error "render-fast: Either a single or double kind of ast expected."))))

(define (linearize-contents-list-fast contents-list put tag-name xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table always-render-white-space?)
  (for-each 
    (lambda (contents) (linearize-contents-fast contents put tag-name xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table always-render-white-space?))
    contents-list))


(define (linearize-contents-fast contents put tag-name xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table 
                                 always-render-white-space?)
  (cond ((char-ref? contents) (put (xml-render-char-ref contents)))
        ((cdata? contents) 
          (let ((white-space-printed? #f)) ; true if last printed char is white space
           (if (and xml-transliterate-character-data? (not (member tag-name xml-non-transliteration-elements)))

	       (do ((lgt (string-length contents))
		    (i 0 (+ i 1))
		    )
		   ((= i lgt) 'done)
                 (let* ((ch (html-char-transform (string-ref contents i) xml-char-transformation-table)) ; a string!
                        (ch-white-space? (and (not (empty-string? ch)) (string-of-char-list? ch white-space-char-list)))
                       )
                  (if always-render-white-space?
                      (put ch)
                      (begin
                        (if (not (and white-space-printed? ch-white-space?))
			    (if ch-white-space? (put #\space) (put ch)))
                        (if ch-white-space? (set! white-space-printed? #t) (set! white-space-printed? #f))
                      )
                  )
                 )
               )

               (do ((lgt (string-length contents))
		    (i 0 (+ i 1))
		    )
		   ((= i lgt) 'done)
                 (let ((ch (string-ref contents i)))
                  (if always-render-white-space?
                      (put ch)
                      (begin
                       (if (not (and white-space-printed? (memq ch white-space-char-list)))
			   (put ch))
                       (if (memq ch white-space-char-list) (set! white-space-printed? #t) (set! white-space-printed? #f))
                      )
                  )
                 )
               )
            )
           )
        )
        ((forced-white-space? contents) (put #\space))
        ((ast? contents) (render-fast contents put always-render-white-space?))
   )
)


; Return whether key is a css-key, for instance a key like css:k or CSS:k
; A true value is the key proper, such as k (as a string).
(define (xml-css-key? key)
  (let* ((key-str (symbol->string key))
         (lgt (string-length key-str)))
   (if
    (and (>= lgt 4) 
         (or (eq? (string-ref key-str 0) #\c) (eq? (string-ref key-str 0) #\C))
         (or (eq? (string-ref key-str 1) #\s) (eq? (string-ref key-str 1) #\S))
         (or (eq? (string-ref key-str 2) #\s) (eq? (string-ref key-str 2) #\S))
         (eq? (string-ref key-str 3) #\:))
    (substring key-str 4 lgt)
    #f)))

; Split attribute-alist in two parts: the HTML attributes and the CSS attributes.
; Return (cons html-alist css-alist) where html-alist is the association lists of HTML attributes
; and css-alist is the association list of CSS attributes.
(define (html-css-split attribute-alist html-alist css-alist)
  (cond ((null? attribute-alist) (cons (reverse html-alist) (reverse css-alist)))
        ((xml-css-key? (caar attribute-alist)) 
            (html-css-split (cdr attribute-alist) html-alist (cons (car attribute-alist) css-alist)))
        (else 
            (html-css-split (cdr attribute-alist) (cons (car attribute-alist) html-alist) css-alist))))


; ----------------------------------------------------------------------------------------------------------------------------------------------------------
; Pretty printed, fast AST rendering.
; The procedures pp-render-fast, pp-linearize-contents-list-fast, and pp-linearize-contents-fast
; are very similar to render-fast, linearize-contents-list-fast, and linearize-contents-fast. 
; This is an important notice for maintenance purposes. 
; In case of single line rendering of an XML fragment, pp-render-fast calls render-fast.


(define (pp-render-fast ast put always-render-white-space? start-col single-lining?)
 (if (single-liner-form? ast start-col preferred-maximum-width)
     (render-fast ast put #f)
     (let* ((tag-name (ast-element-name ast)) 
            (contents-list (ast-subtrees ast))
            (attribute-properlist (ast-attributes ast)) ; split in normal and css here
            (attribute-alist (propertylist-to-alist attribute-properlist))
            (reorganized-attribute-alist (html-css-split attribute-alist '() '()))
            (html-attribute-alist (car reorganized-attribute-alist))
            (css-attribute-alist  (cdr reorganized-attribute-alist))
            (kind (ast-kind ast))

            (language (ast-language ast))
            (xml-transliterate-character-data? (xml-transliterate-character-data-in? language))
            (xml-non-transliteration-elements (xml-non-transliteration-elements-in language))
            (xml-preformatted-text-elements (xml-preformatted-text-elements-in language)) ; (list "pre")
            (xml-char-transformation-table (xml-char-transformation-table-in language))

            (attr-width 0)  ; approximative printed attribute width
            (attr-cnt 0)
            (attr-lgt (length attribute-alist))

           )
       (cond ((eq? kind 'single)        ; empty element
              (put #\<) 
              (put tag-name) 

                                        ; Linearize HTML attributes - duplicate code inlined for double elements!
              (for-each 
               (lambda (attr-pair) 
                 (let ((key (car attr-pair))
                       (val (cdr attr-pair)))
                   (put #\space)
                   (put (symbol->string key))
                   (put #\=) (put quote-char)
                   (do ((lgt (string-length val))
                        (i 0 (+ i 1))
                        )
                       ((= i lgt) 'done)
                     (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
                   (put quote-char)
                   (set! attr-width (+ attr-width (string-length (symbol->string key)) (string-length val) 3))
                   (set! attr-cnt (+ 1 attr-cnt)) 
                   (if (and (> (+ attr-width start-col) preferred-maximum-width) (< attr-cnt attr-lgt))
                       (begin (put #\newline) (put-indentation put (+ (string-length tag-name) start-col 1)) (set! attr-width 0))) ; PP!!
                  )
                 )
               html-attribute-alist)

                                        ; Linearize CSS attributes - duplicate code inlined for double elements!
              (if (not (null? css-attribute-alist))
                  (begin
                    (put #\space) 
                    (put "style=") (put quote-char)
                    (for-each
                     (lambda (attr-pair) 
                       (let* ((key (symbol->string (car attr-pair)))
                              (non-cssed-key (substring key 4 (string-length key)))
                              (val (cdr attr-pair)))
                         (put non-cssed-key)
                         (put #\:)
                         (do ((lgt (string-length val))
                              (i 0 (+ i 1))
                              )
                             ((= i lgt) 'done)
                           (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
                         (put #\;))) 
                     css-attribute-alist)
                    (put quote-char)))

              (put "/>"))

             ((eq? kind 'double)        ; non-empty element
              (put #\<) 
              (put tag-name) 

                                        ; Linearize HTML attributes:
              (for-each 
               (lambda (attr-pair) 
                 (let ((key (car attr-pair))
                       (val (cdr attr-pair)))
                   (put #\space)
                   (put (symbol->string key))
                   (put #\=) (put quote-char)
                   (do ((lgt (string-length val))
                        (i 0 (+ i 1))
                        )
                       ((= i lgt) 'done)
                     (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
                   (put quote-char)
                   (set! attr-width (+ attr-width (string-length (symbol->string key)) (string-length val) 3)) 
                   (set! attr-cnt (+ 1 attr-cnt)) 
                   (if (and (> (+ attr-width start-col) preferred-maximum-width) (< attr-cnt attr-lgt))
                       (begin (put #\newline) (put-indentation put (+ (string-length tag-name) start-col 1)) (set! attr-width 0)))
                  )
                 )
               html-attribute-alist)

                                        ; Linearize CSS attributes:
              (if (not (null? css-attribute-alist))
                  (begin
                    (put #\space) 
                    (put "style=") (put quote-char)
                    (for-each
                     (lambda (attr-pair) 
                       (let* ((key (symbol->string (car attr-pair)))
                              (non-cssed-key (substring key 4 (string-length key)))
                              (val (cdr attr-pair)))
                         (put non-cssed-key)
                         (put #\:)
                         (do ((lgt (string-length val))
                              (i 0 (+ i 1))
                              )
                             ((= i lgt) 'done)
                           (put (html-char-transform (string-ref val i) xml-char-transformation-table)))
                         (put #\;))) 
                     css-attribute-alist)
                    (put quote-char)))

              (put #\>)
              (put #\newline) (put-indentation put (+ start-col indentation-delta)) ; PP!!
              (pp-linearize-contents-list-fast 
                contents-list put tag-name 
                xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table
                (or always-render-white-space? (member tag-name xml-preformatted-text-elements))
                (+ start-col indentation-delta) single-lining?)
              (put #\newline) (put-indentation put start-col) ; PP!!
              (put "</")  (put tag-name) (put #\>))

             (error "pp-render-fast: Either a single or double kind of ast expected.")))))

(define (pp-linearize-contents-list-fast contents-list put tag-name xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table 
                                        always-render-white-space? start-col single-lining?)
  (for-each 
    (lambda (contents) (pp-linearize-contents-fast contents put tag-name xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table
                                                   always-render-white-space? start-col single-lining?))
    contents-list))


(define (pp-linearize-contents-fast contents put tag-name xml-transliterate-character-data? xml-non-transliteration-elements xml-char-transformation-table 
                                    always-render-white-space? start-col single-lining?)
  (cond ((char-ref? contents) (put (xml-render-char-ref contents)))
        ((cdata? contents) 
          (let ((white-space-printed? #f)) ; true if last printed char is white space
           (if (and xml-transliterate-character-data? (not (member tag-name xml-non-transliteration-elements)))

               (do ((lgt (string-length contents))   ; With HTML transformations
                    (i 0 (+ i 1))
                    (j 0 (+ j 1))
                    )
                   ((= i lgt) 'done)
                 (let* ((ch (html-char-transform (string-ref contents i) xml-char-transformation-table)) ; a string!
                        (ch-white-space? (and (not (empty-string? ch)) (string-of-char-list? ch white-space-char-list)))
                       )
                  (if always-render-white-space?     
                      (put ch)
                      (begin
                        (if (not (and white-space-printed? ch-white-space?))
                            (if ch-white-space? 
                               (if (> (+ j start-col) preferred-maximum-width)
                                   (begin (put #\newline) (put-indentation put start-col) (set! j 0)) ; PP!!
                                   (put #\space)
                               )
                               (put ch))
                        )
                        (if ch-white-space? (set! white-space-printed? #t) (set! white-space-printed? #f))
                      )
                  )
                 )  
               )

               (do ((lgt (string-length contents))   ; Without HTML transformations - must be tested carefully.
                    (i 0 (+ i 1))
                    (j 0 (+ j 1))
                   )
                   ((= i lgt) 'done)
                 (let* ((ch (string-ref contents i))
                        (ch-white-space? (memq ch white-space-char-list))
                       )
                  (if always-render-white-space?
                      (put ch)
                      (begin
                       (if (not (and white-space-printed? ch-white-space?))
                           (if ch-white-space?
                               (if (> (+ j start-col) preferred-maximum-width)
                                   (begin (put #\newline) (put-indentation put start-col) (set! j 0)) ; PP!!
                                   (put #\space)
                               )                               
                               (put ch))
                       )
                       (if (memq ch white-space-char-list) (set! white-space-printed? #t) (set! white-space-printed? #f))
                      )
                  )
                 )
               )
            )
           )
        )
        ((forced-white-space? contents)
           (put #\newline) (put-indentation put start-col)  ; PP!!
        )
        ((ast? contents)
           (pp-render-fast contents put always-render-white-space? start-col single-lining?)
        )
   )
)

(define (put-indentation put n)
 (do ((i 1 (+ i 1)))
    ((> i n) 'done) 
  (put #\space)))

(define (single-liner-form? ast start-col max-width)
 (let ((width (measure-xml-in-laml-form ast)))
   (<= (+ width start-col) max-width)))


; Approximative measurement of an AST, string, character reference or white space marker.
(define (measure-xml-in-laml-form x)  
 (cond ((string? x) (string-length x))  ; not accurate due to use of the HTML transformation table.
       ((forced-white-space? x) 1)
       ((char-ref? x) 
        (let ((value (char-ref-value x)))
          (cond ((symbol? value) (+ 2 (string-length (symbol->string value))))
                ((number? value) (+ 3 (string-length (number->string value))))
                (else (laml-error "measure-xml-in-laml-form: Error in character reference" x))))
       )
       ((ast? x)
         (let ((tag-name (ast-element-name x))
               (attributes (propertylist-to-alist (ast-attributes x)))
               (content-list (ast-subtrees x))
              )
           (+ (* 2 (string-length tag-name))                       ; tag-names, start and end
              5                                                    ; angle brackets in start and end tag, '/' in end tag.
              (measure-attribute-list attributes)                  ; attributes
              (sum-list
                (map measure-xml-in-laml-form content-list))       ; contents
           )
         )
       )
       (else (laml-error "measure-xml-in-laml-form: Unknown constituent" x))
  )
)

; Measure attribute-list, which is represented on a-list form.
; Approximative measure - does not take css style attributes into special account
(define (measure-attribute-list attribute-alist)
  (sum-list (map measure-attribute attribute-alist)))

; Measure a single attribute, non CSS. Approximative.
(define (measure-attribute key-val)
 (let ((key (car key-val))
       (val (cdr key-val)))
  (+ (string-length (symbol->string key))
     (string-length val)
     4  ; string quotes, '=', and space after.
  )))


; -------------------------------------------------------------------------------------------------------------------------------------------------
; Attribute linearization functions.  Linearizes to a string.
; These functions are NOT used in render-fast (which uses an imperative, 
; more efficient, html-char-table transformed approach).
; The functions are used in other corners of LAML, such as in relation to AST tree transformation, error messaging,
; and pretty printed rendering.

; Return a pair of strings: The head is the string represeting XML attributes.
; The tail is the string representing the CSS attributes (to go into an HTML style attribute).
(define (xml-linearize-attributes attr-list)
 (let ((lgt (length attr-list)))
  ; catch a typical error here: non-symbol key, probably old-fashioned calling convention from html-v1.
  (if (and (>= lgt 1) (not (symbol? (car attr-list))))
      (error (string-append "xml-linearize-attributes: Non-symbol key encountered: " (as-string (car attr-list)) " in attribute list " (as-string attr-list) ". Maybe conversion problem from html-v1.")))
  (xml-linearize-attributes-1 (reverse attr-list) "" "" lgt attr-list)))

; Internal helping operation to xml-linearize-attributes. The parameter whole-attr-list is for error messaging purposes
(define (xml-linearize-attributes-1 attr-list html-attr-string css-attr-string lgt whole-attr-list)
  (cond ((= lgt  0 ) (cons (strip-trailing-characters (list #\space) html-attr-string) css-attr-string))
        ((>= lgt 2) (let* ((val (car attr-list))   ; val and key are reversed at this time
                           (key (cadr attr-list))
                           (css-key (xml-css-key? key))
                         )
                     (cond (css-key                  ; CSS attribute
                              (xml-linearize-attributes-1 
                                (cddr attr-list)
                                html-attr-string 
                                (string-append (xml-linearize-attribute-pair-css val css-key) ";" css-attr-string)
                                (- lgt 2)
                                whole-attr-list))
                           (else                            ; HTML attribute
                             (xml-linearize-attributes-1 
                               (cddr attr-list)
                               (string-append (xml-linearize-attribute-pair-html val key) " " html-attr-string)
                               css-attr-string
                               (- lgt 2)
                               whole-attr-list)))))
        ((< lgt 2) (error (string-append "Xml-Linearize-attributes-1: Called with an odd length attribute list. Not a Lisp property list: " (as-string whole-attr-list))))))


(define (xml-linearize-attribute-pair-html val key)
  (string-append (as-string key) " = " (string-it (as-string val))))

(define (xml-linearize-attribute-pair-css val key)
  (string-append key ": " (as-string val)))

; ---------------------------------------------------------------------------------------------------------------


;;; Element related functions.
;;; This section contains (mostly) higher-order functions that are related to 
;;; the mirrors of the XML elements.  


;; Bind some attributes content elements of element (the first parameter) and return a new, 'modified element function'.
;; With this, some attributes and some content elements are pre-bound to certain values in the modified element.
;; The parameter attributes-and-contents is of the same form as the parameters to a LAML surface mirror functions.
;; In fact, attributes-and-contents is appended to the actual parameters, which are passed to the modified element function.
;; .example (define a-main (modify-element a 'target "main"))
;; .returns an attribute-modified mirror function
;; .internal-references "similar function" "xml-modify-element-prepend"
(define (xml-modify-element element . attributes-and-contents)
  (lambda parameters (apply element (append parameters attributes-and-contents))))

;; A function similar to xml-modify-element, but instead of appending attributes-and-contents to the
;; actual parameters of the modified function, it prepends attributes-and-contents.
;; .internal-references "similar function" "xml-modify-element"
(define (xml-modify-element-prepend element . attributes-and-contents)
  (lambda parameters (apply element (append attributes-and-contents parameters))))


;;; XML in LAML parametrization and abstraction functions.
;;; The functions in this category are higher-order functions which generate functions that obey XML-in-LAML parameter passing rules.
;;; In other words, the function that are generated use the same parameter conventions as the mirror functions of the HTML and XML elements.

;; Generate a function with XML-in-LAML parameter passing rules which sends its input to an ordinary function, f, with positional parameters via a parameter mediator.
;; The first parameter, f, is typically a web-related function with positional parameter correspondence (such as an existing, old-style 'conveninece function').
;; The parameter parameter-mediator is a function that generates a parameter list for f from
;; content-list and the attributes property list, which is produced by the function xml-sort-tag-parameters.
;; Thus, the parameter-mediator function translates the 'new parameter profile' to the 'old one', which is associated with f.
;; The optional procedure parameter-validator! validates the contents-list and the attribute-list by reporting problems via xml-check-error.
;; The optional f-name parameter is a string corresponding to the name of f (used for error message purposes only).
;; .form (xml-in-laml-parametrization f parameter-mediator [parameter-validator! f-name])
;; .parameter f An 'old style' fuction web function with positional parameters. Is applied on the result of parameter-mediator.
;; .parameter parameter-mediator a function of two parameters (contents-list and attribute property list) which transforms the new style parameters to the old style. f is applied on the output of parameter-mediator!
;; .parameter parameter-validator! a procedure of two parameters (contents-list and attribute property list) which validates the input of the generated function.
;; .parameter f-name The name of the generated function. Solely used for error message purposes.
;; .internal-references "applied function" "xml-sort-tag-parameters"
;; .internal-references "similar function" "xml-in-laml-abstraction"
;; .internal-references "useful par. validator" "required-implied-attributes"
;; .misc See the LAML tutorial - authoring with higher-order functions - for additional discussion and examples.
(define (xml-in-laml-parametrization f  parameter-mediator . optional-parameter-list)
 (let ((parameter-validator! (optional-parameter 1 optional-parameter-list (lambda (co at) #t)))
       (f-name (optional-parameter 2 optional-parameter-list "??"))
      )
  (lambda parameters 
    (let* ((ordered-parameters (xml-sort-tag-parameters parameters f-name))
           (content-parameters (car ordered-parameters))
           (attribute-prop-list (cdr ordered-parameters))
          )
      (parameter-validator! content-parameters attribute-prop-list)
      (apply f (parameter-mediator content-parameters attribute-prop-list))))))


;; Generate a function with XML-in-LAML parameter passing rules which sends its input to an XML or HTML mirror function, or to another XML-in-LAML abstraction.
;; The first parameter, f, is supposed to call one or more XML or HTML mirror function, or another XML-in-LAML abstraction.
;; The function f takes two parameters: A content list and an attribute property list.
;; The optional procedure parameter-validator! validates the contents-list and the attribute-list by reporting problems via xml-check-error.
;; The optional f-name parameter is a string corresponding to the name of f (used for error message purposes only).
;; .form (xml-in-laml-abstraction f [parameter-validator! f-name])
;; .internal-references "applied function" "xml-sort-superficially-tag-parameters"
;; .internal-references "similar function" "xml-in-laml-parametrization"
;; .internal-references "useful par. validator" "required-implied-attributes"
;; .parameter f A transformer function of two parameters - contents and the attribute property list
;; .parameter parameter-validator! A checking procedure of two parameters - contents and the attribute property list. Must report errors explicitly by xml-check-error.
;; .parameter f-name The name of abstraction - used for error message purposes
;; .misc See the LAML tutorial - authoring with higher-order functions - for additional discussion and examples.
(define (xml-in-laml-abstraction f  . optional-parameter-list)
 (let ((parameter-validator! (optional-parameter 1 optional-parameter-list (lambda (co at) #t)))
       (f-name (optional-parameter 2 optional-parameter-list "??"))
      )
  (lambda parameters 
    (let* ((ordered-parameters (xml-sort-superficially-tag-parameters parameters f-name))
           (content-parameters (car ordered-parameters))
           (attribute-prop-list (cdr ordered-parameters))
          )
      (parameter-validator! content-parameters attribute-prop-list)
      (f content-parameters attribute-prop-list)))))


;; A higher-order function which returns an attribute checker. Report problems via xml-check-error.
;; Check that all required-attribute-names are present and that the actual attributes are covered by
;; required-attribute-names and implied-attribute-names together.
;; The generated functions can be used as parameter valdiator procedures in xml-in-laml-abstraction and xml-in-laml-parametrization
;; .internal-references "relevant context of use" "xml-in-laml-abstraction" "xml-in-laml-parametrization"
;; .internal-references "error function" "xml-check-error"
;; .parameter required-attribute-names A list of attribute names (symbols) which are required.
;; .parameter implied-attribute-names A list of additional attribute names (symbols) which are allowed, or the list (*) - a singleton list with the symbol * - in the meaning of any attribute.
;; .form (required-implied-attributes required-attribute-names implied-attribute-names [tag-name])
(define (required-implied-attributes required-attribute-names implied-attribute-names . optional-parameter-list)
 (let ((tag-name (optional-parameter 1 optional-parameter-list "??")))
  (lambda (contents attributes)
   (let ((attribute-names (every-second-element attributes)))
    (xml-check-required-attributes! attribute-names required-attribute-names tag-name)
    (if (not (equal? implied-attribute-names (list '*)))
        (xml-check-for-attribute-existence! attribute-names (append required-attribute-names implied-attribute-names) tag-name))
    ))))


;;; AST traversal and attribute getter functions.
;;; The AST traversal functions extract information from AST by means of traversal and searching.
;;; The AST attribute getter functions extract attribute values from the attribute list of an AST.

;; Traverse all nodes of the AST ast-tree, and return those subtrees that satisfy the node-interesting? predicate.
;; Non-AST constituents of the AST (CDATA and white space markers) are not visited during the traversal.
;; The traversal is done in pre-order.
;; In fact, return the transformed ASTs, applying ast-transformer on each returned tree.
;; The higher-order function ast-of-type? generates useful node-interesting? functions.
;; .internal-references "useful as parameter" "ast-of-type?"
(define (traverse-and-collect-all-from-ast ast-tree node-interesting? ast-transformer)
  (cond ((and (terminal-ast-node? ast-tree) (node-interesting? ast-tree)) (list (ast-transformer ast-tree)))
        ((and (terminal-ast-node? ast-tree) (not (node-interesting? ast-tree))) '())
        ((ast? ast-tree)
           (let ((subtree-results
                   (map 
                    (lambda (subtr) (traverse-and-collect-all-from-ast subtr node-interesting? ast-transformer))
                    (ast-subtrees ast-tree))))
             (if (node-interesting? ast-tree)
                 (cons 
                   (ast-transformer ast-tree)
                   (flatten subtree-results))
                 (flatten subtree-results))))
        ((list? ast-tree)
            (flatten 
              (map 
	       (lambda (tr) (traverse-and-collect-all-from-ast tr node-interesting? ast-transformer))
	       ast-tree)))
        (else '())))


;; Traverse the AST ast-tree, and return the first subtree which satisfies node-interesting.
;; Non-AST constituents of the AST (CDATA and white space markers) are not visited during the traversal.
;; The traversal is done in pre-order.
;; Return #f in case no match is found.
;; In fact, return the transformed AST, applying ast-transformer on the returned tree.
;; The higher-order function ast-of-type? generates useful node-interesting? functions.
;; .internal-references "useful as parameter" "ast-of-type?"
(define (traverse-and-collect-first-from-ast ast-tree node-interesting? ast-transformer)
 (call-with-current-continuation
  (lambda (exit)
    (traverse-and-collect-first-from-ast-help ast-tree node-interesting? ast-transformer exit)))
 )


(define (traverse-and-collect-first-from-ast-help ast-tree node-interesting? ast-transformer exit)
  (cond ((and (terminal-ast-node? ast-tree) (node-interesting? ast-tree)) (exit (ast-transformer ast-tree)))
        ((ast? ast-tree)
           (if (node-interesting? ast-tree)
               (exit (ast-transformer ast-tree))
               (for-each
                 (lambda (subtr) (traverse-and-collect-first-from-ast-help subtr node-interesting? ast-transformer exit))
		 (ast-subtrees ast-tree))))
        ((list? ast-tree)
           (for-each
             (lambda (tr) (traverse-and-collect-first-from-ast-help tr node-interesting? ast-transformer exit))
              ast-tree))
        (else #f))
  
  ; We did not find a tree:
  #f)


;; A higher-order function which returns an attribute getter function on ASTs.
;; The returned function takes an AST as parameter.
;; .parameter attribute-name the attribute name - a symbol.
;; .parameter default-value the default value returned if there is no attribute named attribute-name in the attribute-list.
;; .form (attribute-getter attribute-name [default-value])
(define (attribute-getter attribute-name . optional-parameter-list)
 (let ((default-value (optional-parameter 1 optional-parameter-list #f)))
   (if default-value
    (lambda (ast)
      (let ((attribute-plist (ast-attributes ast)))
        (defaulted-get-prop attribute-name attribute-plist default-value)))
    (lambda (ast)
      (let ((attribute-plist (ast-attributes ast)))
        (get-prop attribute-name attribute-plist))))))

; -----------------------------------------------------------------------------------------------------------------------------
;;; Other useful functions. 
;;; In this section there are non-mirror functions which are useful in in the context of XML and LAML.



;; Convert an abtract syntax tree to a parse tree.
;; Abstract syntax trees are produced by the validating mirror functions.
;; Parse trees are used as an internal format in the HTML and the XML pretty printing procedures.
;; You can use the function pretty-print-html-parse-tree on the parse tree returned by ast-to-parse-tree.
;; .pre-condition The LAML tool html-support (for parsing and pretty printing) must be loaded for this function to work.
;; .misc To load html-support: <kbd>(laml-tool-load "xml-html-support/html-support.scm")</kbd>
(define (ast-to-parse-tree ast)
 (let ((pt (ast-to-parse-tree-1 ast)))
   (make-final-parse-tree 'html-tree (list pt))))


; The function doing the real work in the ast to parse tree conversion.
(define (ast-to-parse-tree-1 ast)
 (letrec 
  ((subtree-transform     ; the function doing the overall ast to parse tree transformation
    (lambda (x) 
     (cond ((ast? x) (ast-to-parse-tree-1 x))
	   ((cdata? x) x)
           ((char-ref? x) (xml-render-char-ref x))
	   ((forced-white-space? x) " ")
	   (else (laml-error "subtree-transform:" "Unknown subtree constituent" (as-string x)))))
    )

   (explicit-space-splicing  ; join explicit spaces, " ", to the left string constituent - to avoid empty lines after pretty printing.
     (lambda (parse-tree)
       (if (tree-entry? parse-tree)
           (let* ((node (root-of-parse-tree parse-tree))
                  (subtrees (subtrees-of-parse-tree parse-tree))
                  (new-subtrees (explicit-space-splicing-lst subtrees '()))
                 )
             (make-parse-tree node new-subtrees))
           parse-tree)
      ))

   (explicit-space-splicing-lst
     (lambda (subtree-list res-lst)
       (cond ((null? subtree-list)
                  (reverse res-lst))
             ((null? (cdr subtree-list))   ; only a single element left - get finished
                (explicit-space-splicing-lst (cdr subtree-list) (cons (car subtree-list) res-lst)))
             ((and (string? (car subtree-list)) (string? (cadr subtree-list)))   ; interesting case - there are two elements
                (if (equal? (cadr subtree-list) " ")    ; join car and cadr via string-append
                    (explicit-space-splicing-lst (cddr subtree-list) (cons (string-append (car subtree-list) (cadr subtree-list)) res-lst))
                    (explicit-space-splicing-lst (cdr subtree-list) (cons (car subtree-list) res-lst))))
             (else (explicit-space-splicing-lst (cdr subtree-list) (cons (car subtree-list) res-lst))))))

   (split-attribute-list  ; split in html and css attributes
     (lambda (attr-lst) (split-attribute-list-1 attr-lst '() '())))

   (split-attribute-list-1  ; iterative helping function
     (lambda (attr-lst html-attr-list css-attr-list)
       (cond ((null? attr-lst) (cons (reverse html-attr-list) (reverse css-attr-list)))
             ((xml-css-key? (car attr-lst)) 
                (split-attribute-list-1 (cddr attr-lst) html-attr-list (cons (cadr attr-lst) (cons (car attr-lst) css-attr-list))))
             (else 
                (split-attribute-list-1 (cddr attr-lst) (cons (cadr attr-lst) (cons (car attr-lst) html-attr-list)) css-attr-list)))))

   (to-css-string     ; linearize all css attributes to a single string
     (lambda (css-attr-list)
       (let ((css-attr-alist (propertylist-to-alist css-attr-list)))
        (string-append
         (list-to-string 
          (map 
           (lambda (css-key-val) (xml-linearize-attribute-pair-css (cdr css-key-val) (xml-css-key? (car css-key-val))))
           css-attr-alist)
          ";") ";"))))

   )
   (let* ((element-name (ast-element-name ast))
 	  (subtrees (ast-subtrees ast))
	  (attr-lst (ast-attributes ast))  ; prop list format
          (attr-list-normal-css (split-attribute-list attr-lst))
          (html-attr-list (car attr-list-normal-css)) ; prop list format
          (css-attr-list (cdr attr-list-normal-css)) ; prop list format
          (attr-lst-result 
            (if (null? css-attr-list)
                html-attr-list
                (cons 'style (cons (to-css-string css-attr-list) html-attr-list))))
	  (kind (ast-kind ast))
   
	 )
     (cond ((eq? kind 'single)
              (make-tag-structure
	       'start-end
	       element-name 
	       attr-lst-result))
           ((eq? kind 'double)
              (explicit-space-splicing
               (make-parse-tree 
		(make-tag-structure
		 'start
		 element-name 
		 attr-lst-result)
		(map subtree-transform subtrees))))
           (else (laml-error "ast-to-parse-tree-1:" "Unknown kind of ast" kind))))))


